"""
Functions that process data before plotting
"""

import math
import numpy as np

from obspy import UTCDateTime
from sohstationviewer.conf import constants as const


maxInt = 1E100
maxFloat = 1.0E100

# TODO: put this in DB
MassPosVoltRanges = {"regular": [0.5, 2.0, 4.0, 7.0],
                     "trillium": [0.5, 1.8, 2.4, 3.5]}
MassPosColorPallets = {"B": ["C", "G", "Y", "R", "M"],
                       "W": ["B", "B", "B", "B", "B"]}


def getMassposValueColors(rangeOpt, chan, cMode, errors, retType='str'):
    if rangeOpt.lower() not in MassPosVoltRanges.keys():
        errors.append("%s: The current selected Mass Position color range "
                      "is '%s' isn't allowable. The accept ranges are:  %s"
                      % (chan, rangeOpt, ', '.join(MassPosVoltRanges.keys())))
        print("ERRORS:", errors)
        return
    massPosVoltRange = MassPosVoltRanges[rangeOpt]
    massPosColorPallet = MassPosColorPallets[cMode]
    mul = -1
    valueColors = []
    """
    remove size for masspos because there are two many points for one plot
    which make the line thicks
    TODO: show something to let user know that it is clickable.
    """
    for i in range(len(massPosVoltRange)):
        if i % 2 == 0:
            mul += 1
        if retType == 'str':
            valueColors.append(
                "%s:%s" % (massPosVoltRange[i], massPosColorPallet[i]))
        else:
            valueColors.append((massPosVoltRange[i], massPosColorPallet[i]))
        if i == len(massPosVoltRange) - 1:
            if retType == 'str':
                valueColors.append(
                    "%s:+%s" % (massPosVoltRange[i], massPosColorPallet[i+1]))
            else:
                valueColors.append(
                    (massPosVoltRange[i], massPosColorPallet[i + 1]))
    if retType == 'str':
        return '|'.join(valueColors)
    return valueColors


def formatTime(time, dateMode, timeMode=None):
    """
    :param parent: parent GUI to display tracking info
    :param time: time to be format, can be UTCDateTime or epoch time
    :param dateMode: the format of date
    :param timeMode: the format of time
    :return: the formated time string
    """
    if isinstance(time, UTCDateTime):
        t = time
    else:
        t = UTCDateTime(time)

    # https://docs.python.org/3/library/datetime.html#
    # strftime-and-strptime-format-codes
    format = ''
    if dateMode == 'YYYY-MM-DD':
        format = '%Y-%m-%d'
    elif dateMode == 'YYYYMMDD':
        format = '%Y%m%d'
    elif dateMode == 'YYYY:DOY':
        format = '%Y:%j'
    if timeMode == 'HH:MM:SS':
        format += " %H:%M:%S"

    ret = t.strftime(format)
    return ret


def getTitle(staID, minTime, maxTime, dateMode):
    """
    :param setID: (netID, statID, locID)
    :param plottingData: a ditionary including:
        { gaps: [(t1,t2),(t1,t2),...]   (in epoch time)
          channels:[cha: {netID, statID, locID, chanID, times, data,
                          startTmEpoch, endTmEpoch}
          earliestUTC: the earliest time of all channels
          latestUTC: the latest time of all channels
    :return: title for the plot
    """
    diff = maxTime - minTime
    hours = diff/3600
    return ("Station: %s  %s  to  %s  (%s)" %
            (staID,
             formatTime(minTime, dateMode, "HH:MM:SS"),
             formatTime(maxTime, dateMode, "HH:MM:SS"),
             round(hours, 2))
            )


def getGaps(gaps, gapMin):
    """
    :param gaps: list of gaps
    :param gapMin: minimum of gaps count in minutes
    return list of gaps of which gaps smaller than gapMin have been removed
    """
    gapMinSec = gapMin * 60
    return [g for g in gaps if (g[1] - g[0]) >= gapMinSec]


def getTimeTicks(earliest, latest, dateFormat, labelTotal):
    """
    split time range into parts to use for tick labels
    Ex: getTimeTicks(1595542860.0, 1595607663.91, 'YYYY-MM-DD', 3)
    :param earliest: earliest epoch time
    :param latest: latest epoch time
    :param dateFormat: YYYY:DOY, YYYY-MM-DD or YYYYMMMDD
    :param labelTotal: number of time label to be displayed,
     others will show as ticks oly
    :return:
        times: list of times
        majorTimes: list of time of lables to be displayed
        majorTimelabels: list of labels displayed
    """
    timeRange = latest - earliest
    if timeRange >= 2592000.0:
        mode = "DD"
        # Time of the nearest midnight before the earliest time.
        time = (earliest // 86400.0) * 86400
        interval = 864000.0
    elif timeRange >= 864000.0:
        mode = "D"
        time = (earliest // 86400.0) * 86400
        interval = 86400.0
    elif timeRange >= 3600.0:
        mode = "H"
        # Nearest hour.
        time = (earliest // 3600.0) * 3600
        interval = 3600.0
    elif timeRange >= 60.0:
        mode = "M"
        # Nearest minute.
        time = (earliest // 60.0) * 60
        interval = 60.0
    else:
        mode = "S"
        time = (earliest // 1)
        interval = 1.0
    times = []
    timeLabels = []
    time += interval
    while time < latest:
        times.append(time)
        timeLabel = formatTime(time, dateFormat, 'HH:MM:SS')
        if mode == "DD" or mode == "D":
            timeLabel = timeLabel[:-9]
        elif mode == "H":
            timeLabel = timeLabel[:-3]
        elif mode == "M" or mode == "S":
            timeLabel = timeLabel
        timeLabels.append(timeLabel)
        time += interval

    ln = len(timeLabels)
    d = math.ceil(len(timeLabels) / labelTotal)
    majorTimes = [times[i] for i in range(ln) if i % d == 0]
    majorTimelabels = [timeLabels[i] for i in range(ln) if i % d == 0]
    return times, majorTimes, majorTimelabels


def getDayTicks():
    times = list(range(const.NO_5M_1H, const.NO_5M_DAY, const.NO_5M_1H))
    majorTimes = list(range(4*const.NO_5M_1H,
                            const.NO_5M_DAY,
                            4*const.NO_5M_1H))
    majorTimeLabels = ["%02d" % int(t/const.NO_5M_1H) for t in majorTimes]
    return times, majorTimes, majorTimeLabels


def getUnitBitweight(chanDB, bitweightOpt, isWF=False):
    """
    :param chanDB: channel's info got from database
    :param bitweightOpt: qpeek's OPTBitWeightRVar
        Commands: Show Seismic Data In Counts - None
                  Use Q330 Low Gain - Low
                  Use Q330 High Gain - High
    :return: unit with fixed point decimal and bitweight if applicable
    """
    plotType = chanDB['plotType']
    # Not all channels use/have this field.
    unit = chanDB['unit']
    try:
        fixPoint = chanDB['fixPoint']
    except Exception:
        fixPoint = 0
    unitBitweight = ''

    if plotType in ['linesDots', 'linesSRate', 'linesMasspos']:

        if fixPoint == 0:
            unitBitweight = "{}%s" % unit
        else:
            unitBitweight = "{:.%sf}%s" % (fixPoint, unit)
    if chanDB['channel'] == 'SEISMIC':
        if fixPoint != 0:
            unitBitweight = "{:.%sf}%s" % (fixPoint, unit)
        else:
            if bitweightOpt in ["low", "high"]:
                unitBitweight = "{}V"
            else:
                unitBitweight = "{}%s" % unit
    return unitBitweight


def convertWFactor(cData, convertFactor):
    """
    convertFactor = 150mV/count = 150V/1000count
    => unit data * convertFactor= data *150/1000 V
    convert data in numpy arrays to a reduced list
    according to max number of datapoint needed
    """
    cData['decData'] = np.multiply(cData['data'], [convertFactor])
    return cData
