import os
import re
from datetime import datetime
from obspy import UTCDateTime


def validateFile(path2file, fileName):
    if fileName.strip() == '.DS_Store' or fileName.startswith('._'):
        # skip mac's info file
        return False

    if not os.path.isfile(os.path.join(path2file)):
        return False
    return True


def displayTrackingInfo(trackingBox, text, type='info'):
    if trackingBox is None:
        print(f"{type}: {text}")
        return

    msg = {'text': text}
    if type == 'error':
        msg['color'] = 'white'
        msg['bgcolor'] = '#e46269'
    elif type == 'warning':
        msg['color'] = '#ffd966'
        msg['bgcolor'] = 'orange'
    else:
        msg['color'] = 'blue'
        msg['bgcolor'] = 'white'
    htmlText = """<body>
        <div style='color:%(color)s; background-color:%(bgcolor)s'>
            %(text)s
        </div>
        </body>"""
    trackingBox.setHtml(htmlText % msg)
    # parent.update()
    trackingBox.repaint()


def getDirSize(dir):
    totalSize = 0
    totalFile = 0
    for path, subdirs, files in os.walk(dir):
        for fileName in files:
            if not validateFile(path, fileName):
                continue
            fp = os.path.join(path, fileName)
            # print("file %s: %s" % (fp, os.path.getsize(fp)))
            totalSize += os.path.getsize(fp)
            totalFile += 1
    return totalSize, totalFile


def getTime6(timeStr):
    """
    Get time from 6 parts string.Ex: 01:251:09:41:35:656/2001:251:09:41:35:656
    """
    year = timeStr.split(':')[0]
    if len(year) == 2:
        return getTime6_2y(timeStr)
    else:
        return getTime6_4y(timeStr)


def getTime6_2y(timeStr):
    """
    Get time from 6 parts string.Ex: 01:251:09:41:35:656
    """
    # pad 0 so the last part has 6 digits to match with the format str
    timeStr = timeStr.ljust(22, "0")
    time = datetime.strptime(timeStr, "%y:%j:%H:%M:%S:%f")
    utcTime = UTCDateTime(time)
    return utcTime.timestamp, time.year


def getTime6_4y(timeStr):
    """
    Get time from 6 parts string.Ex: 2001:251:09:41:35:656
    """
    # pad 0 so the last part has 6 digits to match with the format str
    timeStr = timeStr.ljust(24, "0")
    time = datetime.strptime(timeStr, "%Y:%j:%H:%M:%S:%f")
    utcTime = UTCDateTime(time)
    return utcTime.timestamp, time.year


def getTime4(timeStr, trackingYear, yAdded):
    """
    Get time from 4 parts string. Ex: 253:19:41:42
    """
    if not yAdded:
        # first day => move to next year
        doy = timeStr.split(':')
        if doy == 1:
            trackingYear += 1
            yAdded = True
    timeStr = f'{str(trackingYear)}:{timeStr}'
    time = datetime.strptime(timeStr, "%Y:%j:%H:%M:%S")
    utcTime = UTCDateTime(time)
    return utcTime.timestamp, time.year, yAdded


def getVal(text):
    """
    return value part including +/-, remove str that follows
    """
    REVal = '^\+?\-?[0-9]+\.?[0-9]?'            # noqa: W605
    return float(re.search(REVal, text).group())


def isBinaryStr(text):
    return lambda b: bool(b.translate(None, text))


######################################
# BEGIN: rtnPattern(In, Upper = False)
# LIB:rtnPattern():2006.114 - Logpeek
def rtnPattern(text, upper=False):
    """
    return format of the string with:
     + 0 for digit
     + a for lowercase
     + A for upper case
     + remain special character
    """
    rtn = ""
    for c in text:
        if c.isdigit():
            rtn += "0"
        elif c.isupper():
            rtn += "A"
        elif c.islower():
            rtn += "a"
        else:
            rtn += c
    # So the A/a chars will always be A, so the caller knows what to look for
    if upper is True:
        return rtn.upper()
    return rtn
# END: rtnPattern


####################
# BEGIN: fmti(Value)
# LIB:fmti():2018.235
#   Just couldn't rely on the system to do this, although this won't handle
#   European-style numbers.
def fmti(Value):
    Value = int(Value)
    if Value > -1000 and Value < 1000:
        return str(Value)
    Value = str(Value)
    NewValue = ""
    # There'll never be a + sign.
    if Value[0] == "-":
        Offset = 1
    else:
        Offset = 0
    CountDigits = 0
    for i in range(len(Value) - 1, -1 + Offset, -1):
        NewValue = Value[i] + NewValue
        CountDigits += 1
        if CountDigits == 3 and i != 0:
            NewValue = "," + NewValue
            CountDigits = 0
    if Offset != 0:
        if NewValue.startswith(","):
            NewValue = NewValue[1:]
        NewValue = Value[0] + NewValue
    return NewValue
# END: fmti
