
import os
import math
from struct import unpack

import numpy as np
from obspy.core import Stream, read as read_ms
from PySide2 import QtWidgets

from sohstationviewer.model.mseed.blockettes_reader import (
    readNextBlkt, ReadBlocketteError)
from sohstationviewer.conf.dbSettings import dbConf
from sohstationviewer.conf import constants as const
from sohstationviewer.model.reftek.from_rt2ms import core


def readSOHMSeed(path2file, fileName,
                 SOHStreams, logData, netsProbInFile, trackInfo):
    """
    Read ms and add trace in self.streams to merge later
    Or read log wrap in ms and add to logData under channel in ms header
    """
    stream = read_ms(path2file)

    file = None
    nets = set()
    stats = set()
    netStats = set()
    chanIDs = set()
    newnetID = None
    for trace in stream:
        netID = trace.stats['network'].strip()
        if netID != "None":
            nets.add(netID)
        if len(nets) > 1:
            for netProb in netsProbInFile:
                netProbSet = set(netProb)
                if nets.issubset(netProbSet) or netProbSet.issubset(nets):
                    newnetID = netsProbInFile[netProb]
                    if nets != netProbSet:
                        netsProbInFile[tuple(nets)] = newnetID
                    break
            if newnetID is None:
                msg = (f"There are more than one net in file {fileName}.\n"
                       "Please select one.")
                msgBox = QtWidgets.QMessageBox()
                msgBox.setText(msg)
                netButtons = []
                _nets = sorted(list(nets))
                for net in _nets:
                    netButtons.append(msgBox.addButton(
                        net, QtWidgets.QMessageBox.ActionRole))
                msgBox.exec_()

                selectedIdx = netButtons.index(msgBox.clickedButton())
                newnetID = _nets[selectedIdx]
                netsProbInFile[tuple(nets)] = newnetID

    for trace in stream:
        if newnetID is not None:
            # use one net ID for all traces in a file
            trace.stats['network'] = newnetID
        netID = trace.stats['network'].strip()
        chanID = trace.stats['channel'].strip()
        staID = trace.stats['station'].strip()
        stats.add(staID)
        chanIDs.add(chanID)
        netStats.add((netID, staID))
        if trace.stats.mseed['encoding'] == 'ASCII':
            file = readASCII(
                path2file, file, staID, chanID, trace, logData, trackInfo)
        else:
            if staID not in SOHStreams:
                SOHStreams[staID] = {}
            if chanID not in SOHStreams[staID]:
                SOHStreams[staID][chanID] = Stream()
            SOHStreams[staID][chanID].append(trace)

    if file is not None:
        file.close()
    return {'nets': list(nets),
            'stats': sorted(list(stats)),
            'netStats': sorted(list(netStats)),
            'chanIDs': sorted(list(chanIDs))
            }


def readSOHTrace(trace):
    tr = {}
    tr['chanID'] = trace.stats['channel']
    tr['samplerate'] = trace.stats['sampling_rate']
    tr['startTmEpoch'] = trace.stats['starttime'].timestamp
    tr['endTmEpoch'] = trace.stats['endtime'].timestamp
    """
    trace time start with 0 => need to add with epoch starttime
    times and data have type ndarray
    """
    tr['times'] = trace.times() + trace.stats['starttime'].timestamp
    tr['data'] = trace.data
    return tr


def readMPTrace(trace):
    tr = {}
    tr['chanID'] = trace.stats['channel']
    tr['samplerate'] = trace.stats['sampling_rate']
    tr['startTmEpoch'] = trace.stats['starttime'].timestamp
    tr['endTmEpoch'] = trace.stats['endtime'].timestamp
    """
    trace time start with 0 => need to add with epoch starttime
    times and data have type ndarray
    """
    tr['times'] = trace.times() + trace.stats['starttime'].timestamp
    # TODO: MP only has 4 different values, data can be simplified if too big
    tr['data'] = np.round_(trace.data / 3276.7, 1)
    return tr


def readWaveformTrace(trace, staID, chanID, tracesInfo, tmpDir):
    """
    read data from Trace and save data to files to save mem for processing
    since waveform data are big
    """
    # gaps for SOH only for now
    tr = {}
    tr['samplerate'] = trace.stats['sampling_rate']
    tr['startTmEpoch'] = trace.stats['starttime'].timestamp
    tr['endTmEpoch'] = trace.stats['endtime'].timestamp
    """
    trace time start with 0 => need to add with epoch starttime
    times and data have type ndarray
    """
    times = trace.times() + trace.stats['starttime'].timestamp
    data = trace.data
    tr['size'] = times.size

    trIdx = len(tracesInfo)
    tr['times_zf'] = tr['times_f'] = saveData2File(
        tmpDir, 'times', staID, chanID, times, trIdx, tr['size'])
    tr['data_zf'] = tr['data_f'] = saveData2File(
        tmpDir, 'data', staID, chanID, data, trIdx, tr['size'])
    return tr


def readWaveformMSeed(path2file, fileName, staID, chanID,
                      tracesInfo, dataTime, tmpDir):
    """
    Read ms and add trace in self.streams to merge later
    Or read log wrap in ms and add to logData under channel in ms header
    """
    stream = read_ms(path2file)
    for trace in stream:
        tr = readWaveformTrace(trace, staID, chanID, tracesInfo, tmpDir)
        dataTime[0] = min(tr['startTmEpoch'], dataTime[0])
        dataTime[1] = max(tr['endTmEpoch'], dataTime[1])
        tracesInfo.append(tr)


def readWaveformReftek(rt130, staID, readData, dataTime, tmpDir):
    stream = core.Reftek130.to_stream(
        rt130,
        headonly=False,
        verbose=False,
        sort_permuted_package_sequence=True)
    for trace in stream:
        chanID = trace.stats['channel']
        samplerate = trace.stats['sampling_rate']
        if chanID not in readData:
            readData[chanID] = {
                "tracesInfo": [],
                "samplerate": samplerate}
        tracesInfo = readData[chanID]['tracesInfo']
        tr = readWaveformTrace(trace, staID, chanID, tracesInfo, tmpDir)
        dataTime[0] = min(tr['startTmEpoch'], dataTime[0])
        dataTime[1] = max(tr['endTmEpoch'], dataTime[1])
        tracesInfo.append(tr)


def readASCII(path2file, file, staID, chanID, trace, logData, trackInfo):
    # TODO: read this in MSEED HEADER
    byteorder = trace.stats.mseed['byteorder']
    h = trace.stats
    logText = "\n\n**** STATE OF HEALTH: "
    logText += ("From:%s  To:%s\n" % (h.starttime, h.endtime))
    textFromData = trace.data.tobytes().decode()
    logText += textFromData
    if textFromData != '':
        logText += textFromData
    else:
        recLen = h.mseed['record_length']
        if file is None:
            file = open(path2file, 'rb')
        databytes = file.read(recLen)
        followingBlktsTotal = unpack('%s%s' % (byteorder, 'B'),
                                     databytes[39:40])[0]
        if followingBlktsTotal > 1:
            nextBlktByteNo = 48 + 8  # header + blkt1000(SEED info)
        else:
            nextBlktByteNo = 0
        while nextBlktByteNo != 0:
            try:
                nextBlktByteNo, info = readNextBlkt(
                    nextBlktByteNo, databytes, byteorder)
                logText += info
            except ReadBlocketteError as e:
                trackInfo(f"{staID} - {chanID}: {e.msg}", 'error')

    if staID not in logData:
        logData[staID] = {}
    if h.channel not in logData[staID]:
        logData[staID][h.channel] = []
    logData[staID][h.channel].append(logText)
    return file


def readText(path2file, fileName, textLogs, ):
    """
    Read log file and add to logData under channel TEXT
    """
    with open(path2file, 'r') as file:
        try:
            content = file.read()
        except UnicodeDecodeError:
            raise Exception("Can't process file: %s" % fileName, 'error')

        logText = "\n\n** STATE OF HEALTH: %s\n" % fileName
        logText += content
        textLogs.append(logText)

        return True


def saveData2File(tmpDir, tm_data, staID, chanID,
                  tr, trIdx, trSize, postfix=''):
    """
    save time/data to file to free memory for processing
    :param tmpDir: the temp dir to save file in
    :param tm_data: "times"/"data"
    :param staID: station id
    :param chaID: channel id
    :param tr: numpy array of trace data
    :param trIdx: trace index
    :param trSize: trace size
    :param postfix: in case of decimating for zooming,
        add Z at the end of filename
    """
    memFileName = os.path.join(tmpDir,
                               f"{staID}-{chanID}-{tm_data}-{trIdx}{postfix}")
    # print("memFileName:%s - %s" % (memFileName, trSize))
    memFile = np.memmap(memFileName, dtype='int64', mode='w+',
                        shape=trSize)
    memFile[:] = tr[:]
    del tr
    del memFile
    return memFileName


def checkChan(chanID, reqSOHChans, reqWFChans):
    """
    Check if chanID is required
    return type if pass, False if not required
    """
    ret = checkWFChan(chanID, reqWFChans)
    if ret[0] == 'WF':
        if ret[1]:
            return "WF"
        else:
            return False
    if checkSOHChan(chanID, reqSOHChans):
        return "SOH"
    return False


def checkSOHChan(chanID, reqSOHChans):
    """
    check if chanID is an SOH channel or mass position (mseed channel)
    check if it required by user
    no reqSOHChans means read all
    """
    if reqSOHChans == []:
        return True
    if chanID in reqSOHChans:
        return True
    if 'EX?' in reqSOHChans and chanID.startswith('EX'):
        if chanID[2] in ['1', '2', '3']:
            return True
    # if 'VM?' in reqSOHChans and chanID.startswith('VM'):
    if chanID.startswith('VM'):
        # always read mass position
        # TODO: add reqMPChans
        if chanID[2] in ['0', '1', '2', '3', '4', '5', '6']:
            return True
    return False


def checkWFChan(chanID, reqWFChans):
    """
    Check if chanID is a waveform data and is required by user
    TODO: check with more wild card reqWFChans
    """
    wf = ''
    hasChan = False
    if dbConf['seisRE'].match(chanID):
        wf = 'WF'
    if reqWFChans == ['*']:
        hasChan = True
    if chanID in reqWFChans:
        hasChan = True
    return wf, hasChan


def sortData(dataDict):
    """
    Sort data in 'tracesInfo' according to 'startTmEpoch'
    """
    for staID in dataDict:
        for chanID in dataDict[staID]['readData']:
            tracesInfo = dataDict[staID]['readData'][chanID]['tracesInfo']
            tracesInfo = sorted(
                tracesInfo, key=lambda i: i['startTmEpoch'])


def squash_gaps(gaps):
    """
    :param gaps: list of gaps: (start, end, length)
    :return: squased_gaps: all related gaps are squashed extending to
        min start and max end
    """
    squashed_gaps = []
    sgap_indexes = []
    for idx, g in enumerate(gaps):
        if idx in sgap_indexes:
            continue
        squashed_gaps.append(g)
        sgap_indexes.append(idx)
        for idx_, g_ in enumerate(gaps):
            if idx_ in sgap_indexes:
                continue
            if ((g[0] <= g_[0] <= g[1]) or (g[0] <= g_[1] <= g[1])
                    or (g_[0] <= g[0] <= g_[1]) or (g_[0] <= g[1] <= g_[1])):
                sgap_indexes.append(idx_)
                squashed_gaps[-1][0] = min(g[0], g_[0])
                squashed_gaps[-1][1] = max(g[1], g_[1])
            else:
                break
    return squashed_gaps


def downsample(times, data, rq_points):
    if times.size <= rq_points:
        return times, data
    dataMax = max(abs(data.max()), abs(data.min()))
    dataMean = abs(data.mean())
    indexes = np.where(
        abs(data - data.mean()) >
        (dataMax - dataMean) * const.CUT_FROM_MEAN_FACTOR)
    times = times[indexes]
    data = data[indexes]
    # return constant_rate(times, data, rq_points)
    return chunk_minmax(times, data, rq_points)


def constant_rate(times, data, rq_points):
    if times.size <= rq_points:
        return times, data
    rate = int(times.size/rq_points)
    if rate == 1:
        return times, data

    indexes = np.arange(0, times.size, rate)
    times = times[indexes]
    data = data[indexes]

    return times, data


def chunk_minmax(times, data, rq_points):
    x, y = times, data
    final_points = 0
    if x.size <= rq_points:
        final_points += x.size
        return x, y

    if rq_points < 2:
        return np.empty((1, 0)), np.empty((1, 0))

    # Since grabbing the min and max from each
    # chunk, need to div the requested number of points
    # by 2.
    size = rq_points // 2
    cs = math.ceil(times.size / size)

    if cs * size > times.size:
        cs -= 1
        # Length of the trace is not divisible by the number of requested
        # points. So split into an array that is divisible by the requested
        # size, and an array that contains the excess. Downsample both,
        # and combine. This case gives slightly more samples than
        # the requested sample size, but not by much.
        x0 = times[:cs * size]
        y0 = data[:cs * size]

        x1 = times[cs * size:]
        y1 = data[cs * size:]

        dx0, dy0 = downsample(x0, y0, rq_points)

        # right-most subarray is always smaller than
        # the initially requested number of points.
        dx1, dy1 = downsample(x1, y1, cs)

        dx = np.zeros(dx0.size + dx1.size)
        dy = np.zeros(dy0.size + dy1.size)

        # print(dx0.size + dx1.size)

        dx[:dx0.size] = dx0
        dy[:dy0.size] = dy0

        dx[dx0.size:] = dx1
        dy[dy0.size:] = dy1
        del x0
        del y0
        del x1
        del y1
        del times
        del data
        return dx, dy

    x = x.reshape(size, cs)
    y = y.reshape(size, cs)

    imin = np.argmin(y, axis=1)
    imax = np.argmax(y, axis=1)

    rows = np.arange(size)

    mask = np.zeros(shape=(size, cs), dtype=bool)
    mask[rows, imin] = True
    mask[rows, imax] = True

    dx = x[mask]
    dy = y[mask]
    return dx, dy


def trim_downsample_SOHChan(chan, startTm, endTm, firsttime):
    """
    trim off non-included time from chan[orgTrace], downsample, and save to
    chan[times], chan[data] and [logIdx] if the key exist
    """
    # TODO, add logIdx to downsample if using reftex
    # zoom in to the given time
    tr = chan['orgTrace']

    indexes = np.where((startTm <= tr['times']) & (tr['times'] <= endTm))
    chan['times'], chan['data'] = downsample(
        tr['times'][indexes], tr['data'][indexes],
        const.CHAN_SIZE_LIMIT)


def trim_downsample_WFChan(chan, startTm, endTm, firsttime):
    """
    trim off all chans with non included time
    if totalSize of the rest chans > RECAL_SIZE_LIMIT => need to be downsampled
    Read data from tr's filename, downsample the data
    Return data for plotting
    """
    if 'fulldata' in chan:
        # data is small, already has full in the first trim
        return
    # zoom in to the given range
    chan['startIdx'] = 0
    chan['endIdx'] = len(chan['tracesInfo'])

    if ((startTm > chan['tracesInfo'][-1]['endTmEpoch']) or
            (endTm < chan['tracesInfo'][0]['startTmEpoch'])):
        return False

    indexes = [index for index, tr in enumerate(chan['tracesInfo'])
               if tr['startTmEpoch'] > startTm]
    if indexes != []:
        chan['startIdx'] = indexes[0]
        if chan['startIdx'] > 0:
            chan['startIdx'] -= 1  # startTm in middle of trace
    else:
        chan['startIdx'] = 0

    indexes = [idx for (idx, tr) in enumerate(chan['tracesInfo'])
               if tr['endTmEpoch'] <= endTm]
    if indexes != []:
        chan['endIdx'] = indexes[-1]
        if chan['endIdx'] < len(chan['tracesInfo']) - 1:
            chan['endIdx'] += 1     # endTm in middle of trace
    else:
        chan['endIdx'] = 0
    chan['endIdx'] += 1  # a[x:y+1] = [a[x], ...a[y]

    zTracesInfo = chan['tracesInfo'][chan['startIdx']:chan['endIdx']]
    totalSize = sum([tr['size'] for tr in zTracesInfo])
    if not firsttime and totalSize > const.RECAL_SIZE_LIMIT:
        # size too big, recalc only when it's small enough
        return
    try:
        del chan['times']
        del chan['data']
    except Exception:
        pass
    rq_points = 0
    if totalSize > const.CHAN_SIZE_LIMIT:
        rq_points = int(const.CHAN_SIZE_LIMIT / len(zTracesInfo))
    elif firsttime:
        chan['fulldata'] = True

    chan['times'] = []
    chan['data'] = []
    for trIdx, tr in enumerate(zTracesInfo):
        times = np.memmap(tr['times_f'],
                          dtype='int64', mode='r',
                          shape=tr['size'])
        data = np.memmap(tr['data_f'],
                         dtype='int64', mode='r',
                         shape=tr['size'])
        indexes = np.where((startTm <= times) & (times <= endTm))
        times = times[indexes]
        data = data[indexes]
        if rq_points != 0:
            times, data = downsample(times, data, rq_points)
        chan['times'].append(times)
        chan['data'].append(data)

    chan['times'] = np.hstack(chan['times'])
    chan['data'] = np.hstack(chan['data'])
    return


def get_eachDay5MinList(startTm, endTm):
    exactDayTm = (startTm // const.SEC_DAY) * const.SEC_DAY
    exactDayTmList = []

    if startTm < exactDayTm:
        exactDayTmList = [exactDayTm - const.SEC_DAY]

    while exactDayTm < endTm:
        exactDayTmList.append(exactDayTm)
        exactDayTm += const.SEC_DAY

    # list of start/end 5m in each day: eachDay5MinList
    for idx, startDayTm in enumerate(exactDayTmList):
        aDay5Min = np.arange(startDayTm,
                             startDayTm + const.SEC_DAY,
                             const.SEC_5M)
        if idx == 0:
            eachDay5MinList = np.array([aDay5Min])
        else:
            eachDay5MinList = np.vstack((eachDay5MinList, aDay5Min))
    return eachDay5MinList


def get_trimTPSData(chan, startTm, endTm, eachDay5MinList):
    # preset all 0 for all 5 minutes for each day
    tpsData = np.zeros((len(eachDay5MinList), const.NO_5M_DAY))

    # zoom in to the given range
    chan['startIdx'] = 0
    chan['endIdx'] = len(chan['tracesInfo'])

    if ((startTm > chan['tracesInfo'][-1]['endTmEpoch']) or
            (endTm < chan['tracesInfo'][0]['startTmEpoch'])):
        return False

    indexes = [index for index, tr in enumerate(chan['tracesInfo'])
               if tr['startTmEpoch'] > startTm]
    if indexes != []:
        chan['startIdx'] = indexes[0]
        if chan['startIdx'] > 0:
            chan['startIdx'] -= 1  # startTm in middle of trace
    else:
        chan['startIdx'] = 0

    indexes = [idx for (idx, tr) in enumerate(chan['tracesInfo'])
               if tr['endTmEpoch'] <= endTm]
    if indexes != []:
        chan['endIdx'] = indexes[-1]
        if chan['endIdx'] < len(chan['tracesInfo']) - 1:
            chan['endIdx'] += 1     # endTm in middle of trace
    else:
        chan['endIdx'] = 0
    chan['endIdx'] += 1  # a[x:y+1] = [a[x], ...a[y]

    zTracesInfo = chan['tracesInfo'][chan['startIdx']:chan['endIdx']]

    spr = chan['samplerate']
    chan['tps_data'] = []

    startTPSTm = 0
    accDataList = []

    for trIdx, tr in enumerate(zTracesInfo):
        times = np.memmap(tr['times_f'],
                          dtype='int64', mode='r',
                          shape=tr['size'])
        data = np.memmap(tr['data_f'],
                         dtype='int64', mode='r',
                         shape=tr['size'])
        startIndex = 0
        if trIdx == 0:
            # get index of times with closet value to startTm
            startIndex = np.abs(times - startTm).argmin()
            startTPSTm = times[startIndex]

        # identify index in case of overlaps or gaps
        index = np.where((eachDay5MinList <= times[startIndex]) &
                         (eachDay5MinList + const.SEC_5M > times[startIndex]))
        currRow = index[0][0]
        currCol = index[1][0]
        nextTPSTm = startTPSTm + const.SEC_5M
        while endTm >= nextTPSTm:
            nextIndex = int(startIndex + spr * const.SEC_5M)
            if nextIndex >= tr['size']:
                accDataList.append(data[startIndex:tr['size']])
                break
            else:
                accDataList.append(np.square(data[startIndex:nextIndex]))
                accData = np.hstack(accDataList)
                if accData.size == 0:
                    tpsData[currRow, currCol] = 0
                else:
                    tpsData[currRow, currCol] = np.mean(accData)

                startIndex = nextIndex
                currCol += 1
                accDataList = []
                if currCol == const.NO_5M_DAY:
                    currCol = 0
                    currRow += 1
                    nextTPSTm += const.SEC_5M
    chan['tps_data'] = tpsData
    chan['5m_each_day'] = eachDay5MinList


def findTPSTm(given_tm, each_day_5_min_list):
    """
    Find the position of the given time (given_tm) in time-power-squared plot
    :param given_tm: given time
    :param each_day_5_min_list: list of days in which each day includes
                    a list of 288 times with 5 minutes apart
    :return x_idx: index of time in the each_day_5_min_list
    :return y_idx: index of day plotted
                   (look in TimePowerSquaredWidget.getZoomData())
    """
    x_idx = None
    y_idx = None
    for day_idx, day in enumerate(each_day_5_min_list):
        for tm_idx, tm in enumerate(day):
            if tm > given_tm:
                y_idx = - day_idx
                x_idx = tm_idx - 1
                if x_idx < 0:
                    y_idx = -(day_idx - 1)
                    x_idx = len(day) - 1
                break
        if x_idx is not None:
            break
    return x_idx, y_idx


if __name__ == '__main__':
    print(checkChan('HH1', [], []))
