import struct
from sohstationviewer.controller.util import getTime6
from sohstationviewer.model.handling_data import (
    readSOHMSeed, readText, checkChan)


class futils:
    """
    file utilities class
    """

    def __init__(self, infile):
        self.infile = open(infile, 'r+b')

    def close(self):
        self.infile.close()

    def where(self):
        return self.infile.tell()


#########################################################
class FixedHeader:
    """
    mseed fixhdr
    """
    # first 20char
    textblk = timeblk = sampblk = miscblk = []
    textblk = []
    Serial = DHQual = Res = Stat = Chan = Loc = Net = None
    # 20:30
    timeblk = []
    Year = Day = Hour = Min = Sec = Micro = None
    # 30:36
    sampblk = []
    NumSamp = SampFact = SampMult = None
    # 36:48
    miscblk = []
    act = io = DataDHQualFL = numblock = timcorr = bdata = bbblock = None


#########################################################
class MseedHeader(futils):

    def __init__(self, infile):
        """
        initialize file, determine byteorder, sizes, time, and load first fixed
        header
        """
        self.type = self.rate = None
        # try:
        futils.__init__(self, infile)
        # local variable to class
        self.infileseek = self.infile.seek
        self.infilewrite = self.infile.write
        self.infileread = self.infile.read
        self.sunpack = struct.unpack
        self.spack = struct.pack

        self.byteorder = self.ByteOrder()

        if self.byteorder != "unknown":
            # otherwise it might be mseed
            self.type = "mseed"
            # read 1st fixed header in file
            self.fixedhdr()
            (self.filesize, self.blksize) = self.sizes()

        # except Exception as e:
        #     print("e:",e)
        #     pass

    #########################################################

    def ByteOrder(self, seekval=20):
        """
        read file as if it is mseed just pulling time info
        from fixed header and determine if it makes sense unpacked
        as big endian or little endian
        """
        Order = "unknown"
        try:
            # seek to timeblock and read
            self.infileseek(seekval)
            timeblock = self.infileread(10)

            # assume big endian
            (Year, Day, Hour, Min, Sec, junk, Micro) = \
                self.sunpack('>HHBBBBH', timeblock)
            # test if big endian read makes sense
            if 1950 <= Year <= 2050 and \
                    1 <= Day <= 366 and \
                    0 <= Hour <= 23 and \
                    0 <= Min <= 59 and \
                    0 <= Sec <= 59:
                Order = "big"
                self.fmt_order = ">"
            else:
                # try little endian read
                (Year, Day, Hour, Min, Sec, junk, Micro) = \
                    self.sunpack('<HHBBBBH', timeblock)
                # test if little endian read makes sense
                if 1950 <= Year <= 2050 and \
                        1 <= Day <= 366 and \
                        0 <= Hour <= 23 and \
                        0 <= Min <= 59 and \
                        0 <= Sec <= 59:
                    Order = "little"
                    self.fmt_order = "<"
        except Exception:
            pass

        return Order

#########################################################
#
# for blockette descriptions below
# from SEED manual
#
# Field     nbits   Description
# UBYTE     8       Unsigned quantity
# BYTE      8       Two's complement signed quantity
# UWORD     16      Unsigned quantity
# WORD      16      Two's complement signed quantity
# ULONG     32      Unsigned quantity
# LONG      32      Two's complement signed quantity
# CHAR*n    n*8     n characters, each 8 bit and each with
#                   a 7-bit ASCII character (high bit always 0)
# FLOAT     32      IEEE Floating point number
#
# BTIME
#   UWORD   16  Year (e.g. 1987)
#   UWORD   16  J-Day
#   UBYTE   8   Hours of day (0-23)
#   UBYTE   8   Minutes of hour (0-59)
#   UBYTE   8   Seconds of minute (0-59, 60 for leap seconds)
#   UBYTE   8   Unused for data (required for alignment)(
#   UWORD   16  .0001 seconds (0-9999)
#########################################################
    def fixedhdr(self, seekval=0):
        """
        Reads fixed header of 48 bytes (see SEED manual)
        Returns four tuples
                textblk (SeqNum, DHQual, res, Stat, Loc, Chan, Net)
                        Sequence Number (CHAR*6)
                        Data header/quality indicator (CHAR*1)
                        Reserved (CHAR*1)
                        Station identifier code (CHAR*5)
                        Location identifier (CHAR*2)
                        Channel identifier (CHAR*3)
                        Network Code (CHAR*2)
                timeblk (Year, Day, Hour, Min, Sec, junk, Micro)
                        Year (UWORD, 2)
                        Jday (UWORD, 2)
                        Hour (UBYTE, 1)
                        Min (UBYTE, 1)
                        Sec (UBYTE, 1)
                        unused (UBYTE, 1)
                        .0001 seconds (0-9999) (UWORD, 2)
                sampblk (NumSamp, SampFact, SampMult)
                        Number of samples (UWORD, 2)
                        Sample rate factor (see calcrate) (WORD, 2)
                        Sample rate multiplier (see calcrate) (WORD, 2)
                miscblk (act, io, DataDHQualFl, numblock, timcorr, bdata,
                         bblock)
                        Activity flags (UBYTE, 1)
                        I/O and clock flags (UBYTE, 1)
                        Data quality flags (UBYTE, 1)
                        Number of blockettes that follow (UBYTE, 1)
                        Time correction (LONG, 4)
                        Offset to beginning of data (UWORD, 2)
                        Offset to beginning of next blockette (UWORD, 2)
        """
        # local variable
        # self.sunpack=self.sunpack
        try:
            del self.FH
        except Exception:
            pass
        self.FH = FixedHeader()
        try:
            self.infileseek(seekval)
            fhblk = self.infileread(48)
            # station info
            fmtstr = self.fmt_order + "6scc5s2s3s2s"
            self.FH.textblk = self.sunpack(fmtstr, fhblk[0:20])

            # time info
            fmtstr = self.fmt_order + "HHBBBBH"
            self.FH.timeblk = self.sunpack(fmtstr, fhblk[20:30])

            # sample info
            fmtstr = self.fmt_order + "Hhh"
            self.FH.sampblk = self.sunpack(fmtstr, fhblk[30:36])

            # misc info
            fmtstr = self.fmt_order + "BBBBlHH"
            tmpblk = self.sunpack(fmtstr, fhblk[36:48])
            # decompose tmpblk[0-2] into bit fields, create miscblk
            for i in range(3):
                self.FH.miscblk = self.UbytetoStr(tmpblk, i)
                tmpblk = self.FH.miscblk  # recast tmpblk as list

            (self.FH.Serial, self.FH.DHQual, res, self.FH.Stat,
             self.FH.Loc, self.FH.Chan, self.FH.Net) = self.FH.textblk
            (self.FH.Year, self.FH.Day, self.FH.Hour, self.FH.Min,
             self.FH.Sec, junk, self.FH.Micro) = self.FH.timeblk
            (self.FH.NumSamp, self.FH.SampFact,
             self.FH.SampMult) = self.FH.sampblk
            self.rate = self.calcrate()

            return (self.FH.textblk, self.FH.timeblk, self.FH.sampblk,
                    self.FH.miscblk)

        except Exception as e:
            raise Exception("error reading fixed header: %s" % str(e))

    def sizes(self, seekval=0):
        """
        Finds Blockette 1000 and returns file size & Data Record Length
        """
        # try:
        # determine file size
        self.infileseek(0, 2)
        filesize = self.infile.tell()
        # proceed to seekval
        self.infileseek(seekval)
        # self.infileseek(39)
        # assign number of blockettes and offset to next blockette
        nblock = self.FH.miscblk[3]
        nextblock = self.FH.miscblk[6]
        n = 0
        # find blockette 1000
        while n < nblock:
            self.infileseek(nextblock)
            (blktype, newblock) = self.typenxt(nextblock)
            if not blktype:
                return None, None
            if blktype == 1000:
                (type, next, self.encode, order, length,
                 res) = self.blk1000(nextblock)
                return filesize, 2**length
            nextblock = newblock
            n += 1
        # except Exception:
        #     return None, None

    #########################################################

    def UbytetoStr(self, tup, offset):
        """
        converts unsign byte to string values
        """
        list = []
        strval = ""
        # mask bit fields and build string
        for i in range(8):
            mask = 2 ** i
            if tup[offset] & mask:
                strval = "1" + strval
            else:
                strval = "0" + strval

        # build new list with decomposed bit string
        for i in range(len(tup)):
            if i == offset:
                list.append(strval)
            else:
                list.append(tup[i])
        return list

    #########################################################

    def calcrate(self):
        """
        this routine assumes that idread has been called first

        calculate the sample rate of mseed data
        If Sample rate factor > 0 and Sample rate Multiplier > 0,
                rate = Sampfact X Sampmult
        If Sample rate factor > 0 and Sample rate Multiplier < 0,
                rate = -1 X Sampfact/Sampmult
        If Sample rate factor < 0 and Sample rate Multiplier > 0,
                rate = -1 X Sampmult/Sampfact
        If Sample rate factor < 0 and Sample rate Multiplier < 0,
                rate = 1/(Sampfact X Sampmult)
        """
        sampFact = float(self.FH.SampFact)
        sampMult = float(self.FH.SampMult)
        if sampFact > 0 and sampMult > 0:
            rate = sampFact * sampMult
        elif sampFact > 0 and sampMult < 0:
            rate = -1.0 * (sampFact / sampMult)
        elif sampFact < 0 and sampMult > 0:
            rate = -1.0 * (sampMult / sampFact)
        elif sampFact < 0 and sampMult < 0:
            rate = 1.0 / (sampFact * sampMult)
        else:
            rate = sampFact
        return rate

    #########################################################

    def typenxt(self, seekval=0):
        """
        Reads first 4 bytes of blockette
        Returns blockette type and next blockette offset
        """
        try:
            self.infileseek(seekval)
            fmtstr = self.fmt_order + "HH"
            (type, next) = self.sunpack(fmtstr, self.infileread(4))

            # reset back to beginning of blockette
            self.infileseek(-4, 1)
            return type, next
        except Exception:
            return None, None

    #########################################################

    def blk1000(self, seekval=0):
        """
        Data Only SEED Blockette 1000 (8 bytes)
        Returns tuple
                blk
        Blockette type (UWORD, 2)
        Next blockette's byte offset relative to fixed section of header
                (UWORD, 2)
        Encoding Format (BYTE, 1)
        Word Order (UBYTE, 1)
        Data Record Length (UBYTE, 1)
        Reserved (UBYTE, 1)
        """
        self.infileseek(seekval)
        fmtstr = self.fmt_order + "HHbBBB"
        blk = self.sunpack(fmtstr, self.infileread(8))
        return list(blk)

    #########################################################

    def blk1001(self, seekval=0):
        """
        Data Extension Blockette 1001 (8 bytes)
        Returns tuple
                blk
        Blockette type (UWORD, 2)
        Next blockette's byte offset relative to fixed section of header
                (UWORD, 2)
        Timing Quality (UBYTE, 1)
        microsec (UBYTE, 1)
        Reserved (UBYTE, 1)
        Frame count (UBYTE, 1)
        """
        self.infileseek(seekval)
        fmtstr = self.fmt_order + "HHBBBB"
        blk = self.sunpack(fmtstr, self.infileread(8))
        return list(blk)

    #########################################################

    def blk500(self, seekval=0):
        """
        Timing Blockette 500 (200 bytes)
        Returns tuple
                blk
        Blockette type (UWORD, 2)
        Next blockette's byte offset relative to fixed section of header
                (UWORD, 2)
        VCO correction (FLOAT, 4)
        Time of exception (BTime expanded, 10)
        microsec (UBYTE, 1)
        Reception quality (UBYTE, 1)
        Exception count (ULONG, 4)
        Exception type (CHAR*16)
        Clock model (CHAR*32)
        Clock status (CHAR*128)
        """
        self.infileseek(seekval)
        fmtstr = self.fmt_order + "HHf"
        blk1 = self.sunpack(fmtstr, self.infileread(8))
        fmtstr = self.fmt_order + "HHBBBBH"
        timeblk = self.sunpack(fmtstr, self.infileread(10))
        fmtstr = self.fmt_order + "BBI16s32s128s"
        blk2 = self.sunpack(fmtstr, self.infileread(182))

        # incorporate timeblk tuple into blk list
        blk = list(blk1)
        blk.append(timeblk)
        blk = blk + list(blk2)

        return blk

    #########################################################

    def isMseed(self):
        """
        determines if processed file is mseed (return 1) or unknown type
        (return 0)
        """
        # if we don't know byteorder it must not be mseed
        if self.byteorder == "unknown":
            return 0
        else:
            return 1


def readHdrs(path2file, fileName,
             SOHStreams, logData,
             reqSOHChans, reqWFChans,
             netsProbInFile, trackInfo):
    """
    read headers of a given file build dictionary for quick access
    """

    # create object (we've already tested all files for mseed)
    # and get some base info
    # print("fileName:", fileName)
    rdfile = MseedHeader(path2file)
    if rdfile.isMseed():
        try:
            filesize = rdfile.filesize
            blksize = rdfile.blksize
            encode = rdfile.encode
            chanID = rdfile.FH.Chan.strip().decode()
            chanType = checkChan(chanID, reqSOHChans, reqWFChans)
            if not chanType:
                rdfile.close()
                return
        except Exception:
            rdfile.close()
            raise Exception("Cannot determine file and block sizes. File: %s"
                            % fileName)
    else:
        # not Mseed()
        rdfile.close()
        readText(path2file, fileName, logData['TEXT'])
        return

    if chanType == 'SOH':
        readSOHMSeed(path2file, fileName, SOHStreams,
                     logData, netsProbInFile, trackInfo)
        return
    if reqWFChans == []:
        return
    (numblocks, odd_size) = divmod(filesize, blksize)
    nets = set()
    stats = set()
    netStats = set()
    chanIDs = set()
    epochs = []
    startTms = []
    gaps = []
    # looping over total number of blocks in files
    for n in range(numblocks):
        rdfile.fixedhdr(n * blksize)

        chanID = rdfile.FH.Chan.strip().decode()

        chanIDs.add(chanID)
        nets.add(rdfile.FH.Net.strip().decode())
        stats.add(rdfile.FH.Stat.strip().decode())
        netStats.add((rdfile.FH.Net.strip().decode(),
                     rdfile.FH.Stat.strip().decode()))

        t_str = "%d:%03d:%02d:%02d:%02d:%06d" % (
            rdfile.FH.Year, rdfile.FH.Day, rdfile.FH.Hour,
            rdfile.FH.Min, rdfile.FH.Sec, rdfile.FH.Micro)
        startTms.append(t_str)
        startepoch, _ = getTime6(t_str)
        endepoch = None
        if rdfile.rate != 0:
            endepoch = startepoch + rdfile.FH.NumSamp / rdfile.rate

        epochs.append((startepoch, endepoch))

    rdfile.close()

    return {'path2file': path2file,
            'fileName': fileName,
            'nets': list(nets),
            'stats': sorted(list(stats)),
            'netStats': sorted(list(netStats)),
            'chanIDs': sorted(list(chanIDs)),
            'gaps': gaps,
            'encode': encode,
            'startEpoch': epochs[0][0],
            'endEpoch': epochs[-1][1],
            'tracesTotal': len(epochs),
            'spr': rdfile.rate,
            'read': False
            }
