"""
waveform: + indexing and read only the files in selected time,
          + apply saving memory
          + down sample traces before combine data
          + data not use for plotting is kept in files to free memory
          for processing.
SOH: merge all traces (with obspy stream) before down sample.
gaps: for SOH only, can manually calc. gaps for waveform but takes time

"""


import os

from PySide2 import QtWidgets

from sohstationviewer.view.selectbuttonsdialog import SelectButtonDialog
from sohstationviewer.model.data_type_model import DataTypeModel

from sohstationviewer.conf import constants

from sohstationviewer.model.mseed.from_mseedpeek.mseed_header import readHdrs
from sohstationviewer.controller.util import validateFile
from sohstationviewer.model.handling_data import (
    readWaveformMSeed, squash_gaps, checkWFChan,
    sortData, readSOHTrace)


class MSeed(DataTypeModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # keys: 1+ nets     values: net user choose
        self.netsProbInFile = {}
        self.read_soh_and_index_waveform(self.dir)
        self.selectedKey = self.selectStaID()
        if self.selectedKey is None:
            return
        if len(self.reqWFChans) != 0:
            # self.reselectTimeRange(self.selectedKey)
            self.readWFFiles(self.selectedKey)

    def read_soh_and_index_waveform(self, folder):
        """
        read waveform data, soh data, mass position data, gaps from folder
        """
        self.waveformData, soh_stream = self.index_waveform(folder)
        self.SOHData, self.massPosData, self.gaps = self.merge_soh_streams(
            soh_stream)

    def index_waveform(self, folder):
        """
        :param folder: the parent folder for the data set
        :return waveform_data: a dict for waveform data including
        {sta_id: {'filesInfo': {chan_id: {filepath, time range, trace's info}}
                 'readData': { chan_id: {preset info for samplerate
                         and traces_info which is a list of {info and
                         np.memmap's file path of trace data}
                  }
        }}
            (can't use stream of traces since data to large to be merged and
                processed)
        :return soh_streams: a dict of soh channel streams
            {sta_id: {chan_id: mseed stream of traces}}
        """
        soh_streams = {}
        waveform_data = {}
        net_stat_prop = {}
        stat_prop = {}
        chan_prop = {}
        count = 0

        for path, sub_dirs, files in os.walk(folder):
            for file_name in files:
                path2file = os.path.join(path, file_name)
                if not validateFile(path2file, file_name):
                    continue
                count += 1
                if count % 50 == 0:
                    self.trackInfo(
                        f'Read {count} file headers/ SOH files', 'info')

                ret = readHdrs(
                    path2file, file_name, soh_streams, self.logData,
                    self.reqSOHChans, self.reqWFChans,
                    self.netsProbInFile, self.trackInfo)
                if ret is None:
                    continue

                self.nets.update(ret['nets'])
                self.stats.add(ret['stats'][0])

                if len(ret['stats']) > 1:
                    k = tuple(ret['stats'])
                    if k not in stat_prop:
                        stat_prop[k] = []
                    stat_prop[k].append(file_name)
                else:
                    if len(ret['netStats']) > 1:
                        k = tuple(ret['netStats'])
                        if k not in net_stat_prop:
                            net_stat_prop[k] = []
                        net_stat_prop[k].append(file_name)
                if len(ret['chanIDs']) > 1:
                    k = tuple(ret['chanIDs'])
                    if k not in chan_prop:
                        chan_prop[k] = []
                    chan_prop[k].append(file_name)
                self.channels.update(ret['chanIDs'])
                if 'read' not in ret:
                    # not waveform
                    continue

                # TODO currently for waveform not read the file
                #  with 2 stations, 2 channels
                # correctly. Use the first one in the file only
                # When have time will dig more to this
                sta_id = ret['stats'][0]
                chan_id = ret['chanIDs'][0]
                if sta_id not in waveform_data:
                    waveform_data[sta_id] = {"filesInfo": {},
                                             "readData": {}}
                    self.dataTime[sta_id] = [constants.HIGHEST_INT, 0]
                if chan_id not in waveform_data[sta_id]:
                    waveform_data[sta_id]["filesInfo"][chan_id] = []
                    waveform_data[sta_id]["readData"][chan_id] = {
                        "samplerate": ret['spr'],
                        "tracesInfo": []}
                waveform_data[sta_id]['filesInfo'][chan_id].append(ret)

        if len(stat_prop) > 0:
            errmsg = (f"More than one stations in a file: {stat_prop}. "
                      f"Will use the first one.")
            self.trackInfo(errmsg, "error")
        if len(net_stat_prop) > 0:
            errmsg = "More than one netIDs in a file: %s" % net_stat_prop
            self.trackInfo(errmsg, "warning")
        if len(chan_prop) > 0:
            errmsg = (f"More than one channels in a file: {chan_prop} "
                      f"\nThis is a CRITICAL ERROR.")
            self.trackInfo(errmsg, "error")
        return waveform_data, soh_streams

    def merge_soh_streams(self, soh_streams):
        soh_data = {}
        masspos_data = {}
        gaps = {}
        all_gaps = []
        for sta_id in soh_streams:
            self.stats.add(sta_id)
            self.dataTime[sta_id] = [constants.HIGHEST_INT, 0]
            soh_data[sta_id] = {}
            masspos_data[sta_id] = {}
            for chan_id in soh_streams[sta_id]:
                self.channels.add(chan_id)
                stream = soh_streams[sta_id][chan_id]

                stream.merge()
                if len(stream) > 1:
                    nets = [tr.stats['network'].strip() for tr in stream]
                    nets += [f"Combine to {n}" for n in nets]
                    msg = (f"There are more than one net for sta {sta_id}.\n"
                           "Please select one or combine all to one.")
                    msg_box = SelectButtonDialog(message=msg,
                                                 button_labels=nets)
                    msg_box.exec_()
                    sel_net = nets[msg_box.ret]

                    if "Combine" not in sel_net:
                        tr = [tr for tr in stream
                              if tr.stats['network'] == sel_net][0]
                        self.nets.add(sel_net)
                    else:
                        for tr in stream:
                            sel_net = sel_net.replace("Combine to ", "")
                            tr.stats['network'] = sel_net
                            self.nets.add(sel_net)
                        stream.merge()
                        tr = stream[0]
                else:
                    tr = stream[0]

                gaps_in_stream = stream.get_gaps()
                all_gaps += [[g[4].timestamp, g[5].timestamp]
                             for g in gaps_in_stream]
                trace_info = readSOHTrace(tr)
                if chan_id.startswith('VM'):
                    masspos_data[sta_id][chan_id] = {
                        'orgTrace': trace_info}
                else:
                    soh_data[sta_id][chan_id] = {
                        'orgTrace': trace_info}
                self.dataTime[sta_id][0] = min(trace_info['startTmEpoch'],
                                               self.dataTime[sta_id][0])
                self.dataTime[sta_id][1] = max(trace_info['endTmEpoch'],
                                               self.dataTime[sta_id][1])

            gaps[sta_id] = squash_gaps(all_gaps)
        return soh_data, masspos_data, gaps

    def selectStaID(self):
        stats = list(self.stats)
        selectedStaID = stats[0]
        if len(stats) > 1:
            msg = ("There are more than one stations in the given data.\n"
                   "Please select one to one to display")
            msgBox = QtWidgets.QMessageBox()
            msgBox.setText(msg)
            staButtons = []
            for staID in stats:
                staButtons.append(msgBox.addButton(
                    staID, QtWidgets.QMessageBox.ActionRole))
            abortButton = msgBox.addButton(QtWidgets.QMessageBox.Abort)

            msgBox.exec_()

            if msgBox.clickedButton() == abortButton:
                return selectedStaID
            selectedIdx = staButtons.index(msgBox.clickedButton())
            selectedStaID = stats[selectedIdx]
        self.trackInfo(f'Select Station {selectedStaID}', 'info')
        return selectedStaID
    #
    # def reselectTimeRange(self, staID):
    #     """
    #     If there are no time ranges created for the station, create them
    #     Provide a dialog for user to choose a time range. Each button is
    #     a time range. User click one, time range will be changed in
    #     MainWindow (TODO), the dialog will be closed. If necessary,
    #     will change the way to do this later.
    #     """
    #     if staID not in self.timeRanges:
    #         chanID = self.maxTraceTotalChan[staID]
    #         filesInfo = self.waveformData[staID][chanID]['filesInfo']
    #         cutTrace = [
    #             tr for tr in filesInfo
    #             if (self.readStart <= tr['startEpoch'] < self.readEnd) or
    #                (self.readStart < tr['endEpoch'] <= self.readEnd)]
    #         cutTraceTotal = len(cutTrace)
    #         print(f"{chanID } cutTraceTotal: {cutTraceTotal}")
    #         traceTotalDiv = int(
    #             cutTraceTotal / constants.FILE_PER_CHAN_LIMIT)
    #         if traceTotalDiv < 2:
    #             # readStart, readEnd remain unchanged
    #             return
    #         self.timeRanges[staID] = []
    #         FILE_LIM = constants.FILE_PER_CHAN_LIMIT
    #         for i in range(traceTotalDiv):
    #             startTm = cutTrace[i * FILE_LIM]['startEpoch']
    #             if i < (traceTotalDiv - 1):
    #                 endTm = cutTrace[i * FILE_LIM + FILE_LIM - 1]['endEpoch']
    #             else:
    #                 endTm = cutTrace[-1]['endEpoch']
    #             self.timeRanges[staID].append((startTm, endTm))
    #         msg = "Data in the selected time is too big to display.\n"
    #     else:
    #         msg = ""
    #
    #     msg += "Please choose one of the suggested time range below."
    #     tmStrList = []
    #     for tm in self.timeRanges[staID]:
    #         startTm = UTCDateTime(int(tm[0])).strftime("%Y-%m-%d %H:%M")
    #         endTm = UTCDateTime(int(tm[1])).strftime("%Y-%m-%d %H:%M")
    #         tmStrList.append("%s-%s" % (startTm, endTm))
    #     msgBox = SelectButtonDialog(message=msg, buttonLabels=tmStrList)
    #     msgBox.exec_()
    #
    #     self.readStart, self.readEnd = self.timeRanges[staID][msgBox.ret]
    #     self.trackInfo(f"Select time {self.readStart, self.readEnd}", "info")

    def readWFFiles(self, staID):
        count = 0
        for chanID in self.waveformData[staID]['filesInfo']:
            # check chanID
            hasChan = checkWFChan(chanID, self.reqWFChans)
            if not hasChan:
                continue
            tracesInfo = self.waveformData[staID][
                'readData'][chanID]['tracesInfo']

            for fileInfo in self.waveformData[staID]['filesInfo'][chanID]:
                # file have been read
                if fileInfo['read']:
                    continue

                # check time
                hasData = False

                if ((self.readStart <= fileInfo['startEpoch'] <= self.readEnd) or       # noqa: E501
                        (self.readStart <= fileInfo['endEpoch'] <= self.readEnd)):      # noqa: E501
                    hasData = True
                if not hasData:
                    continue
                readWaveformMSeed(fileInfo['path2file'], fileInfo['fileName'],
                                  staID, chanID, tracesInfo,
                                  self.dataTime[staID], self.tmpDir)
                fileInfo['read'] = True
                count += 1
                if count % 50 == 0:
                    self.trackInfo(f'Read {count} waveform files', 'info')

        sortData(self.waveformData)
