
import os
import numpy as np

from PySide2 import QtWidgets
from obspy.core import Stream

from sohstationviewer.model.reftek.from_rt2ms import (
    core, soh_packet, packet)
from sohstationviewer.model.reftek.logInfo import LogInfo
from sohstationviewer.model.data_type_model import DataTypeModel
from sohstationviewer.model.handling_data import (
    readWaveformReftek, squash_gaps, sortData, readMPTrace, readText)

from sohstationviewer.conf import constants

from sohstationviewer.controller.util import validateFile


class RT130(DataTypeModel):
    def __init__(self, *args, **kwarg):
        self.EH = {}
        super().__init__(*args, **kwarg)
        self.keys = set()
        self.reqDSs = self.reqWFChans
        self.massPosStream = {}
        self.readSOH_indexWaveform(self.dir)

        self.selectedKey = self.selectKey()
        if self.selectedKey is None:
            return
        if len(self.reqWFChans) != 0:
            self.readWFFiles(self.selectedKey)

    def readSOH_indexWaveform(self, dir):
        count = 0
        for path, subdirs, files in os.walk(dir):
            for fileName in files:
                path2file = os.path.join(path, fileName)
                if not validateFile(path2file, fileName):
                    continue
                if not self.readReftek130(path2file, fileName):
                    readText(path2file, fileName, self.logData['TEXT'])
                count += 1
                if count % 50 == 0:
                    self.trackInfo(
                        f'Read {count} file headers/ SOH files', 'info')

        self.combineData()

    def selectKey(self):
        self.keys = sorted(list(self.keys))
        selectedKey = self.keys[0]
        if len(self.keys) > 1:
            msg = ("There are more than one keys in the given data.\n"
                   "Please select one to one to display")
            msgBox = QtWidgets.QMessageBox()
            msgBox.setText(msg)
            staButtons = []
            for key in self.keys:
                staButtons.append(msgBox.addButton(
                    key, QtWidgets.QMessageBox.ActionRole))
            abortButton = msgBox.addButton(QtWidgets.QMessageBox.Abort)

            msgBox.exec_()

            if msgBox.clickedButton() == abortButton:
                return selectedKey
            selectedIdx = staButtons.index(msgBox.clickedButton())
            selectedKey = self.keys[selectedIdx]
        self.trackInfo(f'Select Key {selectedKey}', 'info')
        return selectedKey

    def readWFFiles(self, staID):
        count = 0
        for DS in self.waveformData[staID]['filesInfo']:
            readData = self.waveformData[staID]['readData']
            for fileInfo in self.waveformData[staID]['filesInfo'][DS]:
                # file have been read
                if fileInfo['read']:
                    continue

                # check time
                hasData = False
                if ((self.readStart <= fileInfo['startEpoch'] <= self.readEnd) or   # noqa: E501
                        (self.readStart <= fileInfo['endEpoch'] <= self.readEnd)):  # noqa: E501
                    hasData = True
                if not hasData:
                    continue
                readWaveformReftek(fileInfo['rt130'], staID, readData,
                                   self.dataTime[staID], self.tmpDir)
                fileInfo['read'] = True
                count += 1
                if count % 50 == 0:
                    self.trackInfo(f'Read {count} waveform files', 'info')
        sortData(self.waveformData)

    def addLog(self, chan_pkt, logInfo):
        if chan_pkt not in self.logData[self.curKey].keys():
            self.logData[self.curKey][chan_pkt] = []
        self.logData[self.curKey][chan_pkt].append(logInfo)

    def readReftek130(self, path2file, fileName):
        rt130 = core.Reftek130.from_file(path2file)
        unique, counts = np.unique(rt130._data["packet_type"],
                                   return_counts=True)
        nbr_packet_type = dict(zip(unique, counts))

        if b"SH" in nbr_packet_type:
            self.readSH(path2file)
        if b"EH" in nbr_packet_type or b"ET" in nbr_packet_type:
            self.readEHET_MP_indexWF(rt130)
        return True

    def readSH(self, path2file):
        with open(path2file, "rb") as fh:
            str = fh.read()
        data = soh_packet._initial_unpack_packets_soh(str)
        for ind, d in enumerate(data):
            self.curKey = (d['unit_id'].decode(),
                           d['experiment_number'])
            if self.curKey not in self.logData:
                self.logData[self.curKey] = {}
            logs = soh_packet.Packet.from_data(d).__str__()
            # self.addLog(d['packet_type'].decode(), (d['time'], logs))
            if 'SOH' not in self.logData[self.curKey]:
                self.logData[self.curKey]['SOH'] = []
            self.logData[self.curKey]['SOH'].append((d['time'], logs))

    def readEHET_MP_indexWF(self, rt130):
        DS = rt130._data['data_stream_number'][0] + 1
        if DS not in self.reqDSs + [9]:
            return

        ind_EHET = [ind for ind, val in
                    enumerate(rt130._data["packet_type"])
                    if val in [b"EH"]]     # on ly need event header
        nbr_DT_samples = sum(
            [rt130._data[ind]["number_of_samples"]
             for ind in range(0, len(rt130._data))
             if ind not in ind_EHET])

        for ind in ind_EHET:
            d = rt130._data[ind]
            self.curKey = (d['unit_id'].decode(),
                           d['experiment_number'])
            if self.curKey not in self.logData:
                self.logData[self.curKey] = {}
            logs = packet.EHPacket(d).eh_et_info(nbr_DT_samples)
            if 'EHET' not in self.logData[self.curKey]:
                self.logData[self.curKey]['EHET'] = []
            self.logData[self.curKey]['EHET'].append((d['time'], logs))
        if self.curKey not in self.dataTime:
            self.dataTime[self.curKey] = [constants.HIGHEST_INT, 0]
            self.keys.add(self.curKey)
        if DS == 9:
            self.readMassPos(rt130)
        else:
            self.indexWaveForm(rt130, DS)

    def readMassPos(self, rt130):
        if self.curKey not in self.massPosStream:

            self.massPosStream[self.curKey] = Stream()

        stream = core.Reftek130.to_stream(
            rt130,
            headonly=False,
            verbose=False,
            sort_permuted_package_sequence=True)
        for tr in stream:
            self.massPosStream[self.curKey].append(tr)
            self.dataTime[self.curKey][0] = min(
                tr.stats['starttime'].timestamp, self.dataTime[self.curKey][0])
            self.dataTime[self.curKey][1] = max(
                tr.stats['endtime'].timestamp, self.dataTime[self.curKey][1])

    def indexWaveForm(self, rt130, DS):
        if self.curKey not in self.waveformData:
            self.waveformData[self.curKey] = {"filesInfo": {},
                                              "readData": {}}
        if DS not in self.waveformData[self.curKey]["filesInfo"]:
            self.waveformData[self.curKey]["filesInfo"][DS] = []

        stream = core.Reftek130.to_stream(
            rt130,
            headonly=True,
            verbose=False,
            sort_permuted_package_sequence=True)

        self.waveformData[self.curKey]["filesInfo"][DS].append(
            {'rt130': rt130,
             'startEpoch': stream[0].stats['starttime'].timestamp,
             'endEpoch': stream[0].stats['endtime'].timestamp,
             'read': False})

    def combineData(self):
        for k in self.logData:
            if k == 'TEXT':
                continue
            if k not in self.dataTime:
                self.dataTime[k] = [constants.HIGHEST_INT, 0]
                self.keys.add(k)
            self.SOHData[k] = {}
            logs = []
            for pktType in ['SOH', 'EHET']:
                if pktType == 'EHET':
                    logs += '\n\nEvents:'
                try:
                    for log in sorted(self.logData[k][pktType],
                                      key=lambda x: x[0]):
                        # sort log data according to time
                        logs += log[1]
                        if pktType == 'SOH':
                            logs += '\n'
                except KeyError:
                    pass
            logStr = ''.join(logs)
            self.logData[k][pktType] = logStr
            logObj = LogInfo(
                self, self.trackInfo, logStr, k, pktType, self.reqDSs)
            self.dataTime[k][0] = min(logObj.minEpoch, self.dataTime[k][0])
            self.dataTime[k][1] = max(logObj.maxEpoch, self.dataTime[k][1])
            for cName in self.SOHData[k]:
                c = self.SOHData[k][cName]['orgTrace']
                c['times'] = np.array(c['times'])
                c['data'] = np.array(c['data'])
                c['logIdx'] = np.array(c['logIdx'])

        allGaps = []
        self.gaps = {k: [] for k in self.keys}
        self.massPosData = {k: {} for k in self.keys}
        for k in self.massPosStream:
            stream = self.massPosStream[k]
            stream.merge()
            for tr in stream:
                chanID = tr.stats['channel']
                self.massPosData[k][chanID] = {}
                self.massPosData[k][chanID]['orgTrace'] = readMPTrace(tr)
            gaps = stream.get_gaps()
            # gaps is list of [network, station, location, channel, starttime,
            # endtime, duration, number of missing samples]
            allGaps += [[g[4].timestamp, g[5].timestamp] for g in gaps]
            self.gaps[k] = squash_gaps(allGaps)
