from PySide2 import QtWidgets, QtCore

from sohstationviewer.database.proccessDB import (
    executeDB, trunc_addDB, executeDB_dict)
from sohstationviewer.controller.processing import readChannels, detectDataType
from sohstationviewer.controller.util import displayTrackingInfo

INSTRUCTION = """
Place lists of channels to be read in the IDs field.\n
Select the radiobutton for the list to be used in plotting.
"""
TOTAL_ROW = 20

COL = {'sel': 0, 'name': 1, 'dataType': 2, 'IDs': 3, 'clr': 4}


class ChannelPreferDialog(QtWidgets.QWidget):
    def __init__(self, parent, dir_names):
        super(ChannelPreferDialog, self).__init__()
        self.parent = parent
        self.dir_names = dir_names
        self.setWindowTitle("Channel Preferences")
        self.setGeometry(100, 100, 1100, 800)
        main_layout = QtWidgets.QVBoxLayout()
        main_layout.setContentsMargins(7, 7, 7, 7)
        self.setLayout(main_layout)

        main_layout.addWidget(QtWidgets.QLabel(INSTRUCTION))

        self.create_id_table_widget()
        main_layout.addWidget(self.id_table_widget, 1)

        button_layout = self.create_buttons_section()
        main_layout.addLayout(button_layout)

        self.trackingInfoTextBrowser = QtWidgets.QTextBrowser(self)
        self.trackingInfoTextBrowser.setFixedHeight(60)
        main_layout.addWidget(self.trackingInfoTextBrowser)
        self.changed = False

    def create_buttons_section(self):
        h_layout = QtWidgets.QHBoxLayout()
        self.add_db_chan_btn = QtWidgets.QPushButton(
            self, text='Add DB Channels')
        self.add_db_chan_btn.clicked.connect(self.add_db_channels)
        h_layout.addWidget(self.add_db_chan_btn)

        self.scan_chan_btn = QtWidgets.QPushButton(
            self, text='Scan Channels from Data Source')
        self.scan_chan_btn.clicked.connect(self.scan_channels)
        h_layout.addWidget(self.scan_chan_btn)

        self.save_btn = QtWidgets.QPushButton(self, text='Save')
        self.save_btn.clicked.connect(self.save)
        h_layout.addWidget(self.save_btn)

        self.save_add_main_btn = QtWidgets.QPushButton(
            self, text='Save - Add to Main')
        self.save_add_main_btn.clicked.connect(self.save_add_to_main_and_close)
        h_layout.addWidget(self.save_add_main_btn)

        self.close_btn = QtWidgets.QPushButton(self, text='Cancel')
        self.close_btn.clicked.connect(self.close)
        h_layout.addWidget(self.close_btn)
        return h_layout

    def create_id_table_widget(self):
        self.id_table_widget = QtWidgets.QTableWidget(self)
        # self.id_table_widget.verticalHeader().hide()
        self.id_table_widget.setColumnCount(5)
        col_headers = ['', 'Name', 'DataType', 'IDs', 'Clear']
        self.id_table_widget.setHorizontalHeaderLabels(col_headers)
        header = self.id_table_widget.horizontalHeader()
        header.setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        header.setSectionResizeMode(3, QtWidgets.QHeaderView.Stretch)

        self.id_table_widget.setRowCount(TOTAL_ROW)
        self.avail_data_types = self.get_data_types()
        for row_idx in range(TOTAL_ROW):
            self.add_row(row_idx)

        self.curr_row = -1
        self.update_data_table_widget_items()

    @QtCore.Slot()
    def add_row(self, row_idx):
        curr_sel_radio_btn = QtWidgets.QRadioButton(self)
        curr_sel_radio_btn.clicked.connect(
            lambda checked: self.curr_sel_changed(row_idx))
        self.id_table_widget.setCellWidget(
            row_idx, COL['sel'], curr_sel_radio_btn)

        name_line_edit = QtWidgets.QLineEdit(self)
        name_line_edit.textChanged.connect(self.input_changed)
        self.id_table_widget.setCellWidget(
            row_idx, COL['name'], name_line_edit)

        data_type_combo_box = QtWidgets.QComboBox(self)
        data_type_combo_box.currentIndexChanged.connect(self.input_changed)
        data_type_combo_box.addItems(['Unknown'] + self.avail_data_types)
        data_type_combo_box.setCurrentIndex(-1)
        self.id_table_widget.setCellWidget(
            row_idx, COL['dataType'], data_type_combo_box)

        id_line_edit = QtWidgets.QLineEdit(self)
        id_line_edit.textChanged.connect(self.input_changed)
        self.id_table_widget.setCellWidget(
            row_idx, COL['IDs'], id_line_edit)

        del_button = QtWidgets.QPushButton(self, text='CLR')
        del_button.clicked.connect(lambda arg: self.clear_id(row_idx))
        self.id_table_widget.setCellWidget(
            row_idx, COL['clr'], del_button)

    @QtCore.Slot()
    def input_changed(self):
        self.changed = True

    def update_data_table_widget_items(self):
        id_rows = self.get_id_rows()
        # first row
        self.id_table_widget.cellWidget(0, COL['sel']).setChecked(True)
        self.curr_sel_changed(0)
        count = 0
        for r in id_rows:
            self.id_table_widget.cellWidget(
                count, COL['sel']).setChecked(
                True if ['current'] == 1 else False)
            self.id_table_widget.cellWidget(
                count, COL['name']).setText(r['name'])
            self.id_table_widget.cellWidget(
                count, COL['dataType']).setCurrentText(r['dataType'])
            self.id_table_widget.cellWidget(
                count, COL['IDs']).setText(r['IDs'])
            self.id_table_widget.cellWidget(
                count, COL['sel']).setChecked(
                True if r['current'] == 1 else False)
            if ['current'] == 1:
                self.curr_sel_changed(count)
            count += 1
        self.update()

    def get_row(self, row_idx):
        name_widget = self.id_table_widget.cellWidget(row_idx, COL['name'])
        data_type_widget = self.id_table_widget.cellWidget(row_idx,
                                                           COL['dataType'])
        id_widget = self.id_table_widget.cellWidget(row_idx, COL['IDs'])
        clear_widget = self.id_table_widget.cellWidget(row_idx, COL['clr'])
        return name_widget, data_type_widget, id_widget, clear_widget

    @QtCore.Slot()
    def curr_sel_changed(self, row_idx):
        self.curr_row = row_idx
        (self.name_widget, self.data_type_widget,
         self.id_widget, self.clear_widget) = self.get_row(row_idx)
        self.input_changed()

    @QtCore.Slot()
    def clear_id(self, row_idx):
        (name_widget, data_type_widget, id_widget, clear_widget) =\
            self.get_row(row_idx)
        if id_widget.text().strip() != "":
            msg = ("Are you sure you want to delete the channel IDs of  "
                   "row #%s?" % (row_idx + 1))
            result = QtWidgets.QMessageBox.question(
                self, "Confirmation", msg,
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
            if result == QtWidgets.QMessageBox.No:
                return
            self.changed = True
        if name_widget.text() != '':
            self.changed = True
        name_widget.setText('')
        if data_type_widget.currentText != '':
            self.changed = True
        data_type_widget.setCurrentIndex(-1)
        if id_widget.text() != '':
            self.changed = True
        id_widget.setText('')

    def validate_row(self, check_data_type=False):
        if self.curr_row == -1:
            msg = "Please select a row."
            QtWidgets.QMessageBox.information(self, "Select row", msg)
            return False

        if check_data_type:
            self.data_type = self.data_type_widget.currentText()
            if self.data_type not in self.avail_data_types:
                msg = ("Please select a data type that isn't 'Unknown' for "
                       "the selected row.")
                QtWidgets.QMessageBox.information(
                    self, "Select data type", msg)
                return False
        return True

    @QtCore.Slot()
    def add_db_channels(self):
        if not self.validate_row(check_data_type=True):
            return

        db_channels = self.get_db_channels(self.data_type)
        self.id_widget.setText(','.join(db_channels))

    @QtCore.Slot()
    def scan_channels(self):
        if not self.validate_row():
            return

        data_type = detectDataType(self, self.dir_names)
        if data_type in self.avail_data_types:
            self.data_type_widget.setCurrentText(data_type)
        else:
            self.data_type_widget.setCurrenText('Unknown')
        scanned_channels = readChannels(self, self.dir_names)
        self.id_widget.setText(','.join(scanned_channels))

    @QtCore.Slot()
    def save(self):
        if not self.validate_row():
            return
        if self.changed:
            msg = ("All IDs in the database will be overwritten with "
                   "current IDs in the dialog.\nClick Cancel to stop updating "
                   "database.")
            result = QtWidgets.QMessageBox.question(
                self, "Confirmation", msg,
                QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
            if result == QtWidgets.QMessageBox.Cancel:
                return False
        sql_list = []
        for row_idx in range(TOTAL_ROW):
            sql = self.save_row_sql(row_idx)
            if sql is not None:
                sql_list.append(sql)
        if len(sql_list) == 0:
            self.parent.IDs = []
            self.parent.IDsName = ''
            self.parent.data_type = 'Unknown'
            return True

        ret = trunc_addDB('ChannelPrefer', sql_list)
        if ret is not True:
            displayTrackingInfo(self.parent, ret, "error")
        self.parent.IDs = [
            t.strip() for t in self.id_widget.text().split(',')]
        self.parent.IDsName = self.name_widget.text().strip()
        self.parent.IDsDataType = self.data_type_widget.currentText()
        return True

    def save_row_sql(self, row_idx):
        current = 1 if self.id_table_widget.cellWidget(
            row_idx, COL['sel']).isChecked() else 0
        name = self.id_table_widget.cellWidget(
            row_idx, COL['name']).text()
        data_type = self.id_table_widget.cellWidget(
            row_idx, COL['dataType']).currentText()
        id = self.id_table_widget.cellWidget(
            row_idx, COL['IDs']).text()
        if id.strip() == '':
            return
        if name.strip() == '' and id.strip() != '':
            msg = f"Please add Name for row {row_idx}."
            QtWidgets.QMessageBox.information(self, "Missing info", msg)
            return
        return (f"INSERT INTO ChannelPrefer (name, IDs, dataType, current)"
                f"VALUES ('{name}', '{id}', '{data_type}', {current})")

    @QtCore.Slot()
    def save_add_to_main_and_close(self):
        if not self.save():
            return
        self.parent.currIDsNameLineEdit.setText(self.parent.IDsName)
        self.parent.allChanCheckBox.setChecked(False)
        self.close()

    def get_data_types(self):
        data_type_rows = executeDB(
            'SELECT * FROM DataTypes ORDER BY dataType ASC')
        return [d[0] for d in data_type_rows]

    def get_db_channels(self, data_type):
        channel_rows = executeDB(
            f"SELECT channel FROM CHANNELS WHERE dataType='{data_type}' "
            f" ORDER BY dataType ASC")
        return [c[0] for c in channel_rows]

    def get_id_rows(self):
        i_ds_rows = executeDB_dict(
            "SELECT name, IDs, dataType, current FROM ChannelPrefer "
            " ORDER BY name ASC")
        return i_ds_rows
