from PySide2 import QtWidgets, QtGui

from sohstationviewer.database.proccessDB import executeDB


def set_widget_color(widget, changed=False, read_only=False):
    palette = QtGui.QPalette()

    if read_only:
        # grey text
        palette.setColor(QtGui.QPalette.Text, QtGui.QColor(100, 100, 100))
        # light blue background
        palette.setColor(QtGui.QPalette.Base, QtGui.QColor(210, 240, 255))
        widget.setReadOnly(True)
        widget.setPalette(palette)
        return
    else:
        # white background
        palette.setColor(QtGui.QPalette.Base, QtGui.QColor(255, 255, 255))
    if changed:
        # red text
        palette.setColor(QtGui.QPalette.Text, QtGui.QColor(255, 0, 0))
    else:
        try:
            if widget.isReadOnly():
                # grey text
                palette.setColor(
                    QtGui.QPalette.Text, QtGui.QColor(100, 100, 100))
            else:
                # black text
                palette.setColor(QtGui.QPalette.Text, QtGui.QColor(0, 0, 0))
        except AttributeError:
            palette.setColor(QtGui.QPalette.Text, QtGui.QColor(0, 0, 0))
    widget.setPalette(palette)


class Ui_DBInfoDialog(QtWidgets.QWidget):
    def __init__(self, parent, column_headers, col_name, table_name,
                 resize_content_columns=[], requested_columns={},
                 need_data_type_choice=False, check_fk=True):
        self.total_col = len(column_headers)
        self.column_headers = column_headers
        self.resize_content_columns = resize_content_columns
        self.requested_columns = requested_columns
        self.need_data_type_choice = need_data_type_choice
        self.col_name = col_name
        self.table_name = table_name
        self.check_fk = check_fk
        super(Ui_DBInfoDialog, self).__init__()

        self.setGeometry(100, 100, 900, 900)
        main_layout = QtWidgets.QVBoxLayout()
        self.setLayout(main_layout)

        button_layout = self.create_buttons_section()
        main_layout.addLayout(button_layout)

        self.create_data_table_widget()
        main_layout.addWidget(self.data_table_widget, 1)
        if self.table_name != '':
            instruction = ("Background: LIGHT BLUE - Non Editable due to "
                           "FK constrain; "
                           "WHITE - Editable.    "
                           "Text: BLACK - Saved; RED - Not saved")
            # TODO: add question mark button to give instruction

            main_layout.addWidget(QtWidgets.QLabel(instruction))

    def addWidget(self, data_list, row_idx, col_idx, foreign_key=False,
                  choices=None, range=None, field_name=''):
        if data_list is None:
            text = str(row_idx)  # row number
        else:
            text = (str(data_list[row_idx][col_idx - 1])
                    if row_idx < len(data_list) else '')

        if data_list is None:
            widget = QtWidgets.QPushButton(text)
        elif choices is None and range is None:
            widget = QtWidgets.QLineEdit(text)
            if field_name == 'convertFactor':
                # precision=6
                validator = QtGui.QDoubleValidator(0.0, 5.0, 6)
                validator.setNotation(QtGui.QDoubleValidator.StandardNotation)
                widget.setValidator(validator)
                if text == '':
                    widget.setText('1')
        elif choices:
            widget = QtWidgets.QComboBox()
            widget.addItems(choices)
            widget.setCurrentText(text)
        elif range:
            widget = QtWidgets.QSpinBox()
            widget.setMinimum(range[0])
            widget.setMaximum(range[1])
            if text in ["", None, 'None']:
                widget.setValue(range[0])
            else:
                widget.setValue(int(text))

        if data_list is None:
            set_widget_color(widget)
            widget.setFixedWidth(40)
            widget.clicked.connect(
                lambda: self.row_number_clicked(widget))
        elif foreign_key:
            set_widget_color(widget, read_only=True)
        else:
            new = False if row_idx < len(data_list) else True
            set_widget_color(widget, read_only=False, changed=new)
            if choices is None and range is None:
                widget.textChanged.connect(
                    lambda changed_text:
                    self.cell_input_change(changed_text, row_idx, col_idx))
            elif choices:
                widget.currentTextChanged.connect(
                    lambda changed_text:
                    self.cell_input_change(changed_text, row_idx, col_idx))
            elif range:
                widget.valueChanged.connect(
                    lambda changed_text:
                    self.cell_input_change(changed_text, row_idx, col_idx))
        self.data_table_widget.setCellWidget(row_idx, col_idx, widget)

    def row_number_clicked(self, widget):
        self.data_table_widget.selectRow(int(widget.text()))
        self.data_table_widget.repaint()

    def create_buttons_section(self):
        h_layout = QtWidgets.QHBoxLayout()

        if self.need_data_type_choice:
            self.data_type_combo_box = QtWidgets.QComboBox(self)
            data_type_rows = executeDB('SELECT * FROM DataTypes')
            self.data_type_combo_box.addItems([d[0] for d in data_type_rows])
            self.data_type_combo_box.currentTextChanged.connect(
                self.data_type_changed)
            h_layout.addWidget(self.data_type_combo_box)
        if self.table_name != '':
            self.add_row_btn = QtWidgets.QPushButton(self, text='ADD ROW')
            self.add_row_btn.setFixedWidth(300)
            self.add_row_btn.clicked.connect(self.add_new_row)
            h_layout.addWidget(self.add_row_btn)

            self.save_changes_btn = QtWidgets.QPushButton(
                self, text='SAVE CHANGES')
            self.save_changes_btn.setFixedWidth(300)
            self.save_changes_btn.clicked.connect(self.save_changes)
            h_layout.addWidget(self.save_changes_btn)

        self.close_btn = QtWidgets.QPushButton(self, text='CLOSE')
        self.close_btn.setFixedWidth(300)
        self.close_btn.clicked.connect(self.close)
        h_layout.addWidget(self.close_btn)
        return h_layout

    def create_data_table_widget(self):
        self.data_table_widget = QtWidgets.QTableWidget(self)
        self.data_table_widget.verticalHeader().hide()
        self.data_table_widget.setColumnCount(self.total_col)
        self.data_table_widget.setHorizontalHeaderLabels(self.column_headers)
        header = self.data_table_widget.horizontalHeader()
        header.setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        for i in self.resize_content_columns:
            header.setSectionResizeMode(
                i, QtWidgets.QHeaderView.ResizeToContents)

        if self.need_data_type_choice:
            self.data_type = self.data_type_combo_box.currentText()
        self.update_data_table_widget_items()

    def update_data_table_widget_items(self):
        self.data_list = self.get_data_list()
        row_count = len(self.data_list)
        row_count = 1 if row_count == 0 else row_count
        self.data_table_widget.setRowCount(row_count)
        for i in range(len(self.data_list)):
            fk = self.check_data_foreign_key(self.data_list[i][0])
            self.add_row(i, fk)
        if len(self.data_list) == 0:
            """
            No Row, should leave 1 empty row
            """
            self.clearFirstRow()
        self.update()

    def add_new_row(self):
        row_position = self.data_table_widget.rowCount()
        self.data_table_widget.insertRow(row_position)
        self.add_row(row_position)
        self.data_table_widget.scrollToBottom()
        self.data_table_widget.repaint()  # to show row's header
        self.data_table_widget.cellWidget(row_position, 1).setFocus()

    def remove_row(self, remove_row_idx):
        self.data_table_widget.removeRow(remove_row_idx)
        for i in range(remove_row_idx, self.data_table_widget.rowCount()):
            cell_widget = self.data_table_widget.cellWidget(i, 0)
            cell_widget.setText(str(i))

    def cell_input_change(self, changed_text, rowidx, colidx):
        """
        If cell's value is changed, text's color will be red
        otherwise, text's color will be black
        """
        changed = False
        if rowidx < len(self.data_list):
            if changed_text != self.data_list[rowidx][colidx - 1]:
                changed = True
            cell_widget = self.data_table_widget.cellWidget(rowidx, colidx)
            set_widget_color(cell_widget, changed=changed)

    def check_data_foreign_key(self, val):
        if not self.check_fk:
            return False
        sql = (f"SELECT {self.col_name} FROM channels "
               f"WHERE {self.col_name}='{val}'")
        param_rows = executeDB(sql)
        if len(param_rows) > 0:
            return True
        else:
            return False

    def reset_row_inputs(self, reset, widget_idx, list_idx):
        for col_idx in range(1, self.total_col):
            cell_widget = self.data_table_widget.cellWidget(
                widget_idx,
                col_idx
            )
            read_only = False
            if hasattr(cell_widget, 'isReadOnly'):
                read_only = cell_widget.isReadOnly()
            if not read_only:
                if reset == 1:
                    org_val = self.data_list[list_idx][col_idx - 1]
                    if isinstance(cell_widget, QtWidgets.QLineEdit):
                        cell_widget.setText(str(org_val))
                    elif isinstance(cell_widget, QtWidgets.QComboBox):
                        cell_widget.setCurrentText(str(org_val))
                    elif isinstance(cell_widget, QtWidgets.QSpinBox):
                        cell_widget.setValue(int(org_val))
                set_widget_color(cell_widget)

    def add_row(self, row_idx, fk=False):
        pass

    def data_type_changed(self):
        pass

    def save_changes(self):
        try:
            self.data_table_widget.focusWidget().clearFocus()
        except AttributeError:
            pass
        self.remove_count = 0
        self.insert_count = 0
        self.skip_count = 0
        row_count = self.data_table_widget.rowCount()
        for i in range(row_count):
            widget_idx = i - (row_count - self.data_table_widget.rowCount())
            list_idx = (i - self.remove_count
                        + self.insert_count - self.skip_count)
            row_inputs = self.get_row_inputs(widget_idx)
            reset = self.update_data(row_inputs, widget_idx, list_idx)
            if reset > -1:
                self.reset_row_inputs(reset, widget_idx, list_idx)

    def update_data(self, row, widget_idx, list_idx, insert_sql, update_sql):
        """
        update data_table_widget and data_list according to the action to
        add, remove, update
        :param row: values of processed row in data_table_widget
        :param widget_idx: index of the process rows in data_table_widget
        :param list_idx: index of the process rows in self.data_list
        :param col_name: key db column in the table
        :param table_name: data table name
        :param insert_sql: query to insert a row to the table
        :param update_sql: query to update a row in the table
        :return -1 for doing nothing
                0 no need to reset input values to org row
                1 need to reset input values to org row
        """
        if list_idx < len(self.data_list):
            org_row = self.data_list[list_idx]
            if row[0] == "":
                msg = (f"The {self.col_name} of '{org_row}' at "
                       f"row {widget_idx} has been changed to blank. "
                       f"Are you sure you want to delete this row in "
                       f"database?")
                result = QtWidgets.QMessageBox.question(
                    self, "Delete %s?" % org_row, msg,
                          QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No
                )
                if result == QtWidgets.QMessageBox.No:
                    # reset the first widget value and call the function again
                    # to check other fields
                    cell_widget = self.data_table_widget.cellWidget(
                        widget_idx, 1
                    )
                    cell_widget.setText(org_row[0])
                    row[0] = org_row[0]
                    self.update_data(row, widget_idx, list_idx)
                else:
                    sql = (f"DELETE FROM {self.table_name} "
                           f"WHERE {self.col_name}='{org_row[0]}'")
                    executeDB(sql)
                    self.data_list.remove(org_row)
                    self.remove_row(widget_idx)
                    self.remove_count += 1
                    return -1

            if row == org_row:
                return -1
            if (row[0] in [p[0] for p in self.data_list] and
                    self.data_list[list_idx][0] != row[0]):
                msg = (f"The {self.col_name} of {org_row} at row"
                       f" {widget_idx} has been changed to '{row[0]}' "
                       f"which already is in the database. "
                       f"It will be changed back to {org_row[0]}.")
                QtWidgets.QMessageBox.information(self, "Error", msg)
                # reset the first widget value and call the function again
                # to check other fields
                cell_widget = self.data_table_widget.cellWidget(widget_idx, 1)
                cell_widget.setText(org_row[0])
                row[0] = org_row[0]
                self.update_data(row, widget_idx, list_idx)
            else:
                msg = (f"{org_row} at row {widget_idx} has "
                       f"been changed to {row}. Please confirm it.")
                result = QtWidgets.QMessageBox.question(
                    self, "Confirmation", msg,
                    QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel
                )
                if result == QtWidgets.QMessageBox.Cancel:
                    return 1
                else:
                    executeDB(update_sql % org_row[0])
                    self.data_list[list_idx] = row
                    return 0

        if row[0] == "":
            msg = (f"Row {widget_idx} has blank {self.col_name}. "
                   f"It will be removed.")
            QtWidgets.QMessageBox.information(self, "Error", msg)
            self.remove_row(widget_idx)
            return -1
        blank_requested_columns = [self.requested_columns[i]
                                   for i in self.requested_columns.keys()
                                   if row[i - 1] == ""]
        if blank_requested_columns != []:
            msg = (f"Row {widget_idx}: blank "
                   f"{', '.join(blank_requested_columns)} which require some "
                   f"value(s).")
            QtWidgets.QMessageBox.information(self, "Error", msg)
            return -1

        if row[0] in [p[0] for p in self.data_list]:
            msg = (f"The {self.col_name} '{row[0]}' is already in the "
                   f"database. Row {widget_idx} will be removed.")
            QtWidgets.QMessageBox.information(self, "Error", msg)
            self.remove_row(widget_idx)
            return -1
        executeDB(insert_sql)
        self.data_list.append(row)
        self.insert_count += 1
        return 0
