from PySide2 import QtCore, QtWidgets
from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as Canvas)
from matplotlib import pyplot as pl
from matplotlib.patches import ConnectionPatch, Rectangle
from matplotlib.ticker import AutoMinorLocator
import numpy as np

from sohstationviewer.controller.plottingData import (
    getTitle, getGaps, getTimeTicks, getUnitBitweight, getMassposValueColors)
from sohstationviewer.controller.util import displayTrackingInfo, getVal

from sohstationviewer.conf import constants
from sohstationviewer.conf.colorSettings import Clr, set_colors
from sohstationviewer.conf.dbSettings import dbConf

from sohstationviewer.database import extractData

from sohstationviewer.model.handling_data import trim_downsample_SOHChan

BOTTOM = 0.996
BOTTOM_PX = 200


class PlottingWidget(QtWidgets.QScrollArea):
    """
    zorder:
        axis spines: 0
        center line: 1
        lines: 2
        gap, dots, : 3
    """

    def __init__(self, parent, tracking_box, name=''):
        super().__init__()
        self.name = name
        self.parent = parent
        self.tracking_box = tracking_box
        self.peer_plotting_widgets = [self]
        self.plot_no = 0
        self.info_widget = None
        self.widgt = QtWidgets.QWidget(parent)
        self.axes = []
        self.curr_plot_title = None
        self.hidden_plots = {}
        self.zoom_marker1_shown = False
        self.axes = []

        self.width_base = 0.25
        self.width_base_px = 1546
        self.ratio_w = 1

        # width of plotting area
        self.plotting_w = self.width_base
        # X1: 0.19: Width = 20% of 50in (Figure's width)
        # X2: 0.19*2: Width = 40% of 50in
        # X4: 0.19*4: Width = 80% of 50 in
        # height of plotting area
        #     + changed when plots are added or removed
        #     + changed when changing the V-magnify param

        self.plotting_bot = BOTTOM
        # this is the height of a standard plot
        # plot_h = 0.01        # 0.01: Height = 1% of 100in (Figure's height)
        # left of plotting area: no change
        self.plotting_l_base = 0.04
        self.plotting_l_base = self.plotting_l_base
        self.plotting_h = 0.996
        # this is the height of a standard plot
        # plot_h = 0.01        # 0.01: Height = 1% of 100in (Figure's height)
        # left of plotting area: no change
        self.plotting_l = 0.03
        # bottom of plot gap, where we start to draw data
        # self.plotGapB = 0.990
        # distance from left axis to start of label
        self.label_pad = 100
        self.font_size = 7

        """
        Set Figure size 50in width, 100in height.
        This is the maximum size of plotting container.
        add_axes will draw proportion to this size.
        The actual view for user based on size of self.widgt.
        """
        self.fig = pl.Figure(facecolor='white', figsize=(50, 100))
        self.fig.canvas.mpl_connect('pick_event', self.on_pick_on_artist)
        self.fig.canvas.mpl_connect('button_press_event', self.on_pick)

        self.canvas = Canvas(self.fig)
        self.canvas.setParent(self.widgt)
        self.setWidget(self.widgt)
        self.set_colors('B')

    # ============================= EVENT ==============================
    def resizeEvent(self, event):
        # print("resizeEvent")
        geo = self.maximumViewportSize()
        # print("resize geo:", geo)

        # set view size fit with the scroll's view port size
        self.widgt.setFixedWidth(geo.width())
        self.ratio_w = geo.width() / self.width_base_px
        self.plotting_w = self.ratio_w * self.width_base
        self.plotting_l = self.ratio_w * self.plotting_l_base
        if self.plot_no == 0:
            self.widgt.setFixedHeight(geo.height())

        return super(PlottingWidget, self).resizeEvent(event)

    def contextMenuEvent(self, event):
        if self.axes == []:
            return
        context_menu = QtWidgets.QMenu(self)
        remove_plot_act = context_menu.addAction(
            "Remove %s" % self.curr_plot_title)
        remove_plot_act.setStatusTip("Remove the current Plot")
        remove_plot_act.triggered.connect(self.hide_curr_plot)

        if self.hidden_plots != {}:
            show_all_plot_act = context_menu.addAction("Show All Plots")
            show_all_plot_act.triggered.connect(self.show_all_hidden_plots)
        show_plot_acts = []
        for i in sorted(self.hidden_plots.keys()):
            show_plot_acts.append(
                context_menu.addAction('Show hidden Plot %s' % i))
            show_plot_acts[-1].triggered.connect(
                lambda *arg, index=i: self.show_hidden_plot(index))
        context_menu.exec_(self.mapToGlobal(event.pos()))

    def get_timestamp(self, event):
        x, y = event.x, event.y  # coordinate data
        inv = self.axes[0].transData.inverted()
        # convert to timestamp, bc event.xdata is None in the space bw axes
        xdata = inv.transform((x, y))[0]
        return xdata

    def zoom_bw_markers(self, xdata):
        if self.curr_min_x == xdata:
            return
        self.zoom_marker1_shown = False
        [self.curr_min_x, self.curr_max_x] = sorted([self.curr_min_x, xdata])
        self.set_lim()
        self.zoom_marker1.set_visible(False)
        self.zoom_marker2.set_visible(False)

    def on_pick(self, event):
        """
        xmouse, ymouse = event.mouseevent.xdata, event.mouseevent.ydata
        # x, y = artist.get_xdata(), artist.get_ydata()
        # ind = event.ind
        print('Artist picked:', artist)
        # print(self.plots)
        # print('Index:', self.plots.index(event.artist))
        # print('{} vertices picked'.format(len(ind)))
        # print('Pick between vertices {} and {}'.format(
        min(ind), max(ind) + 1))
        print('x, y of mouse: {:.2f},{:.2f}'.format(xmouse, ymouse))
        # print('Data point:', x[ind[0]], y[ind[0]])
                    print("mouseevent:", dir(event.mouseevent))
            print("guievent:", dir(event.guiEvent))
            print("modifiers:", event.guiEvent.modifiers())
        """
        # print(dir(event))
        modifiers = event.guiEvent.modifiers()
        try:
            # on_pick() take place after on_pick_on_artist where
            # tps_t was assigned in TimePowerSquareWidget
            xdata = self.tps_t
        except AttributeError:
            xdata = self.get_timestamp(event)
        for w in self.peer_plotting_widgets:
            if modifiers == QtCore.Qt.ShiftModifier:
                w.on_shift_click(xdata)
            elif modifiers in [QtCore.Qt.ControlModifier,
                               QtCore.Qt.MetaModifier]:
                w.on_ctrl_cmd_click(xdata)

            else:
                if w.zoom_marker1_shown:
                    w.zoom_bw_markers(xdata)
            w.draw()

    def on_ctrl_cmd_click(self, xdata):
        """
        On ctrl/cmd + left click
        + hide zoom_marker1
        + connect ruler to follow mouse
        + show ruler
        """
        self.zoom_marker1.set_visible(False)
        self.zoom_marker1_shown = False
        try:
            self.fig.canvas.mpl_disconnect(self.follower)
        except AttributeError:
            pass
        self.set_ruler_visibled(self.ruler, xdata)

    def on_shift_click(self, xdata):
        """
        On shift + left click
        if there're no zoom marker
            + hide ruler
            + show zoom_marker1, zoom_marker2
            + connect zoom_marker2 to follow mouse
        else:
            + zoom data in between 2 zoomMarkers

        """
        if not self.zoom_marker1_shown:
            self.ruler.set_visible(False)
            self.set_ruler_visibled(self.zoom_marker1, xdata)
            self.curr_min_x = xdata
            self.zoom_marker1_shown = True
            self.set_ruler_visibled(self.zoom_marker2, xdata)
        else:
            self.zoom_bw_markers(xdata)

    def on_pick_on_artist(self, event):
        artist = event.artist
        if isinstance(artist, pl.Axes):
            self.curr_plot = artist
            self.curr_plot_index = self.axes.index(artist)
            self.curr_plot_title = "Plot %s" % self.curr_plot_index

    def set_ruler_visibled(self, ruler, x):
        ruler.set_visible(True)
        ruler.xy1 = (x, 0)
        ruler.xy2 = (x, self.bottom)
        try:
            if ruler == self.zoom_marker2:
                # make zoom_marker2 follow mouse.
                # need to disconnect when state of rulers change
                self.follower = self.fig.canvas.mpl_connect(
                    "motion_notify_event", self.zoom_marker2_follow_mouse)
        except AttributeError:
            pass

    def zoom_marker2_follow_mouse(self, mouseevent):
        xdata = self.get_timestamp(mouseevent)
        self.zoom_marker2.xy1 = (xdata, 0)
        self.zoom_marker2.xy2 = (xdata, self.bottom)
        self.draw()

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_Escape:
            for w in self.peer_plotting_widgets:
                w.set_rulers_invisible()
                w.zoom_marker1_shown = False
                w.draw()
        return super(PlottingWidget, self).keyPressEvent(event)

    def set_rulers_invisible(self):
        self.ruler.set_visible(False)
        self.zoom_marker1.set_visible(False)
        self.zoom_marker2.set_visible(False)

    # ============================= END EVENT ==============================

    def add_timestamp_bar(self, used_height, top=True, has_label=True):
        """
        set the axes to display timestamp_bar on top of the plotting area
        setting axis off to not display at the begining
        Color in qpeek is DClr[T0]
        :return:
        """
        self.plotting_bot -= used_height
        timestamp_bar = self.canvas.figure.add_axes(
            [self.plotting_l, self.plotting_bot, self.plotting_w, 0.00005],
        )
        timestamp_bar.axis('off')
        timestamp_bar.xaxis.set_minor_locator(AutoMinorLocator())
        timestamp_bar.spines['bottom'].set_color(self.display_color['basic'])
        timestamp_bar.spines['top'].set_color(self.display_color['basic'])

        if top:
            labelbottom = False
        else:
            labelbottom = True
        timestamp_bar.tick_params(which='major', length=7, width=2,
                                  direction='inout',
                                  colors=self.display_color['basic'],
                                  labelbottom=labelbottom,
                                  labeltop=not labelbottom)
        timestamp_bar.tick_params(which='minor', length=4, width=1,
                                  direction='inout',
                                  colors=self.display_color['basic'])
        if has_label:
            timestamp_bar.set_ylabel('Hours',
                                     fontweight='bold',
                                     fontsize=self.font_size,
                                     rotation=0,
                                     labelpad=self.label_pad,
                                     ha='left',
                                     color=self.display_color['basic'])

        # self.update_timestamp_bar(timestamp_bar)
        return timestamp_bar

    def update_timestamp_bar(self, timestamp_bar):
        times, major_times, major_time_labels = getTimeTicks(
            self.curr_min_x, self.curr_max_x, self.date_mode,
            self.time_ticks_total
        )
        timestamp_bar.axis('on')
        timestamp_bar.set_yticks([])
        timestamp_bar.set_xticks(times, minor=True)
        timestamp_bar.set_xticks(major_times)
        timestamp_bar.set_xticklabels(major_time_labels,
                                      fontsize=self.font_size + 2)
        timestamp_bar.set_xlim(self.curr_min_x, self.curr_max_x)

    def create_axes(self, plot_b, plot_h, has_min_max_lines=True):
        ax = self.canvas.figure.add_axes(
            [self.plotting_l, plot_b, self.plotting_w, plot_h],
            picker=True
        )

        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_visible(False)
        if has_min_max_lines:
            ax.spines['top'].set_zorder(0)
            ax.spines['bottom'].set_zorder(0)
            ax.spines['top'].set_color(self.display_color['sub_basic'])
            ax.spines['bottom'].set_color(self.display_color['sub_basic'])

        ax.set_yticks([])
        ax.set_xticks([])
        ax.tick_params(colors=self.display_color['basic'],
                       width=0,
                       pad=-2,
                       labelsize=self.font_size)
        ax.patch.set_alpha(0)
        return ax

    def set_axes_info(self, ax, sample_no_list, sample_no_clr_list=None,
                      label=None, info='', y=None, chan_db=None,
                      linked_ax=None):
        if label is None:
            label = chan_db['label']
        title_ver_alignment = 'center'
        # set info undertitle
        if linked_ax is not None:
            info = label
        if info != '':
            ax.text(
                -0.15, 0.2,
                info,
                horizontalalignment='left',
                verticalalignment='top',
                rotation='horizontal',
                transform=ax.transAxes,
                color=self.display_color['sub_basic'],
                size=self.font_size
            )
            title_ver_alignment = 'top'

        if linked_ax is None:
            # set title on left side
            color = self.display_color['plot_label']
            if label.startswith("DEFAULT"):
                color = self.display_color["warning"]
            ax.text(
                -0.15, 0.6,
                label,
                horizontalalignment='left',
                verticalalignment=title_ver_alignment,
                rotation='horizontal',
                transform=ax.transAxes,
                color=color,
                size=self.font_size + 2
            )

        # set samples' total on right side
        # if sample_no_clr_list is None:
        #     sample_no_clr_list = len(sample_no_list) * ['W']
        if len(sample_no_list) == 1:
            ax.sampleLbl = ax.text(
                1.005, 0.5,
                sample_no_list[0],
                horizontalalignment='left',
                verticalalignment='center',
                rotation='horizontal',
                transform=ax.transAxes,
                # color=Clr[sample_no_clr_list[0]],
                color=self.display_color['basic'],
                size=self.font_size
            )
        else:
            # bottom
            ax.sampleLbl = ax.text(
                1.005, 0.25,
                sample_no_list[0],
                horizontalalignment='left',
                verticalalignment='center',
                rotation='horizontal',
                transform=ax.transAxes,
                # color=Clr[sample_no_clr_list[0]],
                color=self.display_color['basic'],
                size=self.font_size
            )
            # top
            ax.sampleLbl = ax.text(
                1.005, 0.75,
                sample_no_list[1],
                horizontalalignment='left',
                verticalalignment='center',
                rotation='horizontal',
                transform=ax.transAxes,
                # color=Clr[sample_no_clr_list[0]],
                color=self.display_color['basic'],
                size=self.font_size
            )

        if y is None:
            # draw center line
            ax.plot([self.curr_min_x, self.curr_max_x],
                    [0, 0],
                    color=self.display_color['sub_basic'],
                    linewidth=0.5,
                    zorder=1
                    )
            ax.spines['top'].set_visible(False)
            ax.spines['bottom'].set_visible(False)
        else:
            try:
                min_y = y.min()
                max_y = y.max()
                ax.spines['top'].set_visible(True)
                ax.spines['bottom'].set_visible(True)
                ax.unit_bw = getUnitBitweight(
                    chan_db, self.parent.bitweightOpt
                )
                self.set_axes_ylim(ax, min_y, max_y)
            except Exception:
                pass

    def set_axes_ylim(self, ax, min_y, max_y):
        min_y = round(min_y, 7)
        max_y = round(max_y, 7)
        if max_y > min_y:
            ax.set_yticks([min_y, max_y])
            ax.set_yticklabels(
                [ax.unit_bw.format(min_y), ax.unit_bw.format(max_y)])
        if min_y == max_y:
            max_y += 1
            ax.set_yticks([min_y])
            ax.set_yticklabels([ax.unit_bw.format(min_y)])
        ax.set_ylim(min_y, max_y)

    def add_gap_bar(self, gaps):
        """
        set the axes to display gap_bar on top of the plotting area
        setting axis off to not display at the beginning
        :return:
        """
        if self.parent.minGap is None:
            return
        self.gaps = getGaps(gaps, float(self.parent.minGap))
        self.plotting_bot -= 0.003
        self.gap_bar = self.create_axes(self.plotting_bot,
                                        0.001,
                                        has_min_max_lines=False)
        self.update_gap_bar()

    def update_gap_bar(self):
        gap_label = "%sm" % self.parent.minGap
        # TODO: calculate gap limit
        # self.gaps = []
        h = 0.001  # height of rectangle represent gap
        self.set_axes_info(self.gap_bar, [len(self.gaps)],
                           label=gap_label)
        # draw gaps
        for i in range(len(self.gaps)):
            x = self.gaps[i][0]
            w = self.gaps[i][1] - self.gaps[i][
                0]  # width of rectangle represent gap
            self.gap_bar.add_patch(
                Rectangle(
                    (x, - h / 2), w, h, color='r', picker=True, lw=0., zorder=3
                )
            )  # on top of center line

    def get_height(self, ratio, bw_plots_distance=0.0015):
        plot_h = 0.0012 * ratio  # ratio with figure height
        self.plotting_bot -= plot_h + bw_plots_distance
        self.plotting_bot_pixel += 19 * ratio
        return plot_h

    # -------------------- Different color dots ----------------------- #
    def plot_none(self):
        """
        plot with nothing needed to show rulers
        """
        plot_h = 0.00001
        bw_plots_distance = 0.0001
        self.plotting_bot -= plot_h + bw_plots_distance
        ax = self.create_axes(self.plotting_bot, plot_h,
                              has_min_max_lines=False)
        ax.x = None
        ax.plot([0], [0], linestyle="")
        return ax

    def plot_multi_color_dots(self, c_data, chan_db, chan, ax, linked_ax):
        """
        plot scattered dots with colors defined by valueColors:
          *:W or -1:_|0:R|2.3:Y|+2.3:G
           with colors: RYGMC in dbSettings.py
           _: not plot
        :param data: data of the channel which is list of (time, value)
        :param chan_db: info of channel from DB
        :return:
        """
        if linked_ax is not None:
            ax = linked_ax
        if ax is None:
            plot_h = self.get_height(chan_db['height'])
            ax = self.create_axes(
                self.plotting_bot, plot_h, has_min_max_lines=False)

        x = []
        prev_val = -constants.HIGHEST_INT

        if chan_db['valueColors'] in [None, 'None', '']:
            chan_db['valueColors'] = '*:W'
        value_colors = chan_db['valueColors'].split('|')
        for vc in value_colors:
            v, c = vc.split(':')
            val = getVal(v)
            if c == '_':
                prev_val = val
                continue

            if v.startswith('+'):
                points = [c_data['times'][i]
                          for i in range(len(c_data['data']))
                          if c_data['data'][i] > val]
            elif v == '*':
                points = c_data['times']
            else:
                points = [c_data['times'][i]
                          for i in range(len(c_data['data']))
                          if prev_val < c_data['data'][i] <= val]
            x += points

            ax.plot(points, len(points) * [0], linestyle="",
                    marker='s', markersize=0.5, zorder=3,
                    color=Clr[c], picker=True, pickradius=3)
            prev_val = val

        total_samples = len(x)

        x = sorted(x)
        self.set_axes_info(ax, [total_samples], chan_db=chan_db,
                           linked_ax=linked_ax)
        if linked_ax is None:
            ax.x = x
        else:
            ax.linkedX = x
        return ax

    # def plotDotsMasspos(self, c_data, chan_db, chan, linked_ax):
    #     valueColors = getMassposValueColors(
    #         self.parent.massPosVoltRangeOpt, chan, self.cMode, self.errors)
    #     if valueColors is None:
    #         return
    #     chan_db['valueColors'] = valueColors
    #     return self.plot_multi_color_dots(c_data, chan_db, chan, linked_ax)

    # ---------------------------- up/down dots ---------------------------- #
    def plot_up_down_dots(self, c_data, chan_db, chan, ax, linked_ax):
        """
        data with 2 different values defined in valueColors
        """
        if linked_ax is not None:
            ax = linked_ax
        if ax is None:
            plot_h = self.get_height(chan_db['height'])
            ax = self.create_axes(
                self.plotting_bot, plot_h, has_min_max_lines=False)

        val_cols = chan_db['valueColors'].split('|')
        points_list = []
        colors = []
        for vc in val_cols:
            v, c = vc.split(':')
            val = getVal(v)

            points = [c_data['times'][i]
                      for i in range(len(c_data['data']))
                      if c_data['data'][i] == val]
            points_list.append(points)
            colors.append(c)

        # down dots
        ax.plot(points_list[0], len(points_list[0]) * [-0.5], linestyle="",
                marker='s', markersize=2, zorder=3,
                color=Clr[colors[0]], picker=True, pickradius=3)
        # up dots
        ax.plot(points_list[1], len(points_list[1]) * [0.5], linestyle="",
                marker='s', markersize=2, zorder=3,
                color=Clr[colors[1]], picker=True, pickradius=3)
        x = points_list[0] + points_list[1]
        x = sorted(x)
        ax.set_ylim(-2, 2)
        self.set_axes_info(ax, [len(points_list[0]), len(points_list[1])],
                           sample_no_clr_list=colors,
                           chan_db=chan_db,
                           linked_ax=linked_ax)
        if linked_ax is None:
            ax.x = x
        else:
            ax.linkedX = x
        return ax

    # ----------------------- dots for times, ignore data------------------- #
    def plot_time_dots(self, c_data, chan_db, chan, ax, linked_ax):
        if linked_ax is not None:
            ax = linked_ax
        if ax is None:
            plot_h = self.get_height(chan_db['height'])
            ax = self.create_axes(self.plotting_bot, plot_h)

        color = 'W'
        if chan_db['valueColors'] not in [None, 'None', '']:
            color = chan_db['valueColors'].strip()
        x = c_data['times']
        self.set_axes_info(ax, [len(x)], chan_db=chan_db, linked_ax=linked_ax)

        ax.myPlot = ax.plot(x, [0] * len(x), marker='s', markersize=1.5,
                            linestyle='', zorder=2,
                            color=Clr[color], picker=True,
                            pickradius=3)
        if linked_ax is None:
            ax.x = x
        else:
            ax.linkedX = x
        return ax

    # ----------------------- lines - one color dots ----------------------- #
    def plot_lines_dots(self, c_data, chan_db, chan, ax, linked_ax, info=''):
        """ L:G|D:W """
        if linked_ax is not None:
            ax = linked_ax
        if ax is None:
            plot_h = self.get_height(chan_db['height'])
            ax = self.create_axes(self.plotting_bot, plot_h)

        x, y = c_data['times'], c_data['data']
        self.set_axes_info(ax, [len(x)], chan_db=chan_db,
                           info=info, y=y, linked_ax=linked_ax)
        colors = {}
        if chan_db['valueColors'] not in [None, 'None', '']:
            color_parts = chan_db['valueColors'].split('|')
            for cStr in color_parts:
                obj, c = cStr.split(':')
                colors[obj] = c

        l_color = 'G'
        has_dot = False
        if 'L' in colors:
            l_color = colors['L']
        if 'D' in colors:
            d_color = colors['D']
            has_dot = True

        if not has_dot:
            ax.myPlot = ax.plot(x, y,
                                linestyle='-', linewidth=0.7,
                                color=Clr[l_color])
        else:
            ax.myPlot = ax.plot(x, y, marker='s', markersize=1.5,
                                linestyle='-', linewidth=0.7, zorder=2,
                                color=Clr[l_color],
                                markerfacecolor=Clr[d_color],
                                picker=True, pickradius=3)
        if linked_ax is None:
            ax.x = x
            ax.y = y
        else:
            ax.linkedX = x
            ax.linkedY = y
        return ax

    def plot_lines_s_rate(self, c_data, chan_db, chan, ax, linked_ax):
        """
        multi-line line seismic, one color, line only,
            can apply bit weights in (get_unit_bitweight())
        """
        if c_data['samplerate'] >= 1.0:
            info = "%dsps" % c_data['samplerate']
        else:
            info = "%gsps" % c_data['samplerate']
        return self.plot_lines_dots(c_data, chan_db, chan, ax, linked_ax,
                                    info=info)

    # ----------------------- lines - multi-color dots --------------------- #
    def plot_lines_mass_pos(self, c_data, chan_db, chan, ax, linked_ax):
        value_colors = getMassposValueColors(
            self.parent.massPosVoltRangeOpt, chan,
            self.c_mode, self.errors, retType='tupleList')

        if value_colors is None:
            return

        if ax is None:
            plot_h = self.get_height(chan_db['height'])
            ax = self.create_axes(self.plotting_bot, plot_h)

        ax.x, ax.y = c_data['times'], c_data['data']
        self.set_axes_info(ax, [len(ax.x)], chan_db=chan_db, y=ax.y)
        ax.myPlot = ax.plot(ax.x, ax.y,
                            linestyle='-', linewidth=0.7,
                            color=self.display_color['sub_basic'],
                            zorder=2)[0]
        colors = [None] * len(ax.y)
        sizes = [0.5] * len(ax.y)
        for i in range(len(ax.y)):
            count = 0
            prev_v = 0
            for v, c in value_colors:
                if count < (len(value_colors) - 1):
                    if prev_v < abs(ax.y[i]) <= v:
                        colors[i] = Clr[c]
                        break
                else:
                    # if abs(ax.y[i]) > v:
                    colors[i] = Clr[c]
                    break
                prev_v = v
                count += 1
        ax.scatter(ax.x, ax.y, marker='s', c=colors, s=sizes, zorder=3)
        return ax

    # ---------------------------------------------------------#

    def add_ruler(self, color):
        ruler = ConnectionPatch(
            xyA=(0, 0),
            xyB=(0, self.bottom),
            coordsA="data",
            coordsB="data",
            axesA=self.timestamp_bar_top,
            axesB=self.timestamp_bar_bottom,
            color=color,
        )
        ruler.set_visible(False)
        self.timestamp_bar_bottom.add_artist(ruler)
        return ruler

    def set_lim(self, org_size=False):
        if not org_size:
            for chanID in self.plotting_data1:
                c_data = self.plotting_data1[chanID]
                self.get_zoom_data(c_data, chanID)
            for chanID in self.plotting_data2:
                c_data = self.plotting_data2[chanID]
                self.get_zoom_data(c_data, chanID)
        self.update_timestamp_bar(self.timestamp_bar_top)
        self.update_timestamp_bar(self.timestamp_bar_bottom)
        if hasattr(self, 'gap_bar'):
            self.gap_bar.set_xlim(self.curr_min_x, self.curr_max_x)
            if not org_size:
                new_gaps = [g for g in self.gaps
                            if (self.curr_min_x <= g[0] <= self.curr_max_x
                                or self.curr_min_x <= g[1] <= self.curr_max_x)]

                # reset total of samples on the right
                self.gap_bar.sampleLbl.set_text(len(new_gaps))
        for ax in self.axes:
            ax.set_xlim(self.curr_min_x, self.curr_max_x)
            if ax.x is None:
                # the plot_none bar is at the end, no need to process
                break
            if not org_size:
                # x, y
                new_x_indexes = np.where((ax.x >= self.curr_min_x) &
                                         (ax.x <= self.curr_max_x))

                # reset total of samples on the right
                ax.sampleLbl.set_text(len(new_x_indexes))

                if len(new_x_indexes) == 0:
                    continue
                if hasattr(ax, 'y'):
                    # don't need to reset y range if ax.y not exist
                    new_x = ax.x[new_x_indexes]
                    new_min_x = min(new_x)
                    new_max_x = max(new_x)
                    try:
                        new_min_x_index = ax.x.index(new_min_x)
                        new_max_x_index = ax.x.index(new_max_x)
                    except AttributeError:
                        new_min_x_index = np.where(ax.x == new_min_x)[0][0]
                        new_max_x_index = np.where(ax.x == new_max_x)[0][0]
                    new_y = ax.y[new_min_x_index:new_max_x_index + 1]
                    new_min_y = min(new_y)
                    new_max_y = max(new_y)
                    self.set_axes_ylim(ax, new_min_y, new_max_y)

    def set_title(self, title, y=100, v_align='top'):
        self.fig.text(-0.15, y, title,
                      verticalalignment=v_align,
                      horizontalalignment='left',
                      transform=self.timestamp_bar_top.transAxes,
                      color=self.display_color['basic'],
                      size=self.font_size)

    def draw(self):
        try:
            self.canvas.draw()
            # a bug on mac:
            # not showing updated info until clicking on another window
            # fix by calling repaint()
            self.widgt.repaint()
        except TypeError:
            pass

    # ######## Functions for outside world #####
    def init_size(self):
        geo = self.maximumViewportSize()
        if self.plot_no == 0:
            # set view size fit with the scroll's view port size
            self.widgt.setFixedWidth(geo.width())
            self.widgt.setFixedHeight(geo.height())

    def set_msg_widget(self, msg_widget):
        self.msg_widget = msg_widget

    def set_background_color(self, color='black'):
        self.fig.patch.set_facecolor(color)
        self.draw()

    def reset_view(self):
        """
        reset all zooms back to the first plotting
        """
        if self.axes == []:
            return
        self.curr_min_x = self.min_x
        self.curr_max_x = self.max_x
        self.set_lim()
        self.draw()

    def clear(self):
        if self.zoom_marker1.get_visible():
            self.zoom_marker1.set_visible(False)
        else:
            self.zoom_marker1.set_visible(True)
        # self.fig.clear()
        # self.axes = []
        self.draw()

    def apply_convert_factor(self, c_data, convert_factor):
        """
        convertFactor = 150mV/count = 150V/1000count
        => unit data * convertFactor= data *150/1000 V
        """
        c_data['data'] = np.multiply(c_data['data'], [convert_factor])

    # =============================== axes ================================
    def plot_channels(self, start_tm, end_tm, key, data_time,
                      gaps, channel_list, time_ticks_total,
                      plotting_data1, plotting_data2):
        """
        :param key: staID for mseed, (device, expNo) for reftek
        :param plottingData: a ditionary including:
            { gaps: [(t1,t2),(t1,t2),...]   (in epoch time)
              channels:{cha: {netID, statID, locID, chanID, times, data,
                              samplerate, startTmEpoch, endTmEpoch}       #
              earliestUTC: the earliest time of all channels
              latestUTC: the latest time of all channels
        :timeTicksTotal: max number of tick to show on time bar

        Data set: {channelname: [(x,y), (x,y)...]
        """
        self.plotting_data1 = plotting_data1
        self.plotting_data2 = plotting_data2
        self.processing_log = []  # [(message, type)]
        self.errors = []
        if self.axes != []:
            self.fig.clear()
        self.date_mode = self.parent.dateFormat.upper()
        self.time_ticks_total = time_ticks_total

        self.min_x = self.curr_min_x = max(data_time[0], start_tm)
        self.max_x = self.curr_max_x = min(data_time[1], end_tm)
        self.plot_no = len(self.plotting_data1) + len(self.plotting_data2)
        title = getTitle(key, self.min_x, self.max_x, self.date_mode)
        self.plotting_bot = BOTTOM
        self.plotting_bot_pixel = BOTTOM_PX
        self.axes = []

        self.timestamp_bar_top = self.add_timestamp_bar(0.003)
        self.set_title(title)
        self.add_gap_bar(gaps)
        not_found_chan = [c for c in channel_list
                          if c not in self.plotting_data1.keys()]
        if len(not_found_chan) > 0:
            msg = (f"The following channels is in Channel Preferences but "
                   f"not in the given data: {not_found_chan}")
            self.processing_log.append((msg, 'warning'))

        for chan_id in self.plotting_data1:
            chan_db = extractData.getChanPlotInfo(chan_id,
                                                  self.parent.dataType)
            if chan_db['height'] == 0:
                # not draw
                continue
            if chan_db['channel'] == 'DEFAULT':
                msg = (f"Channel {chan_id}'s "
                       f"definition can't be found database.")
                displayTrackingInfo(self.tracking_box, msg, 'warning')

            if chan_db['plotType'] == '':
                continue
            self.plotting_data1[chan_id]['chan_db'] = chan_db
            self.get_zoom_data(self.plotting_data1[chan_id], chan_id,
                               self.plotting_data1, True)

        for chan_id in self.plotting_data2:
            chan_db = extractData.getChanPlotInfo(chan_id,
                                                  self.parent.dataType)
            self.plotting_data2[chan_id]['chan_db'] = chan_db
            self.get_zoom_data(self.plotting_data2[chan_id], chan_id, True)

        self.axes.append(self.plot_none())
        self.timestamp_bar_bottom = self.add_timestamp_bar(0.003, top=False)
        self.set_lim(org_size=True)
        self.bottom = self.axes[-1].get_ybound()[0]
        self.ruler = self.add_ruler(self.display_color['time_ruler'])
        self.zoom_marker1 = self.add_ruler(self.display_color['zoom_marker'])
        self.zoom_marker2 = self.add_ruler(self.display_color['zoom_marker'])
        # Set view size fit with the given data
        if self.widgt.geometry().height() < self.plotting_bot_pixel:
            self.widgt.setFixedHeight(self.plotting_bot_pixel)

        self.draw()

    def get_zoom_data(self, c_data, chan_id, plotting_data=None,
                      first_time=False):
        """
        :param setID: (netID, statID, locID)
        :param plotting_data: a ditionary including:
            { gaps: [(t1,t2),(t1,t2),...]   (in epoch time)
              channels:{cha: {netID, statID, locID, chanID, times, data,
                              samplerate, startTmEpoch, endTmEpoch}       #
              earliestUTC: the earliest time of all channels
              latestUTC: the latest time of all channels
        :timeTicksTotal: max number of tick to show on time bar

        Data set: {channelname: [(x,y), (x,y)...]
        """
        chan_db = c_data['chan_db']
        plot_type = chan_db['plotType']
        trim_downsample_SOHChan(c_data, self.curr_min_x, self.curr_max_x,
                                first_time)
        self.apply_convert_factor(c_data, 1)
        if 'ax' not in c_data:
            linked_ax = None
            if chan_db['linkedChan'] not in [None, 'None', '']:
                try:
                    linked_ax = plotting_data['channels'][
                        chan_db['linkedChan']]['ax']
                except KeyError:
                    pass
            ax = getattr(self, dbConf['plotFunc'][plot_type][1])(
                c_data, chan_db, chan_id, None, linked_ax)
            if ax is None:
                return
            c_data['ax'] = ax
            ax.chan = chan_id
            self.axes.append(ax)
        else:
            getattr(self, dbConf['plotFunc'][plot_type][1])(
                c_data, chan_db, chan_id, c_data['ax'], None)

    def hide_plots(self, plot_indexes):
        if self.axes == []:
            return
        plot_indexes = sorted(plot_indexes)
        idx = 0
        total_h = 0
        for i in range(plot_indexes[0], len(self.axes)):
            pos = self.axes[i].get_position()
            pos.y0 += total_h
            pos.y1 += total_h
            if idx < len(plot_indexes) and i == plot_indexes[idx]:
                h = pos.y1 - pos.y0
                total_h += h
                pos.y0 = pos.y1
                idx += 1
                self.hidden_plots[i] = h
            self.axes[i].set_position(pos)

        # currently consider every plot height are all 100px height
        height = self.widgt.geometry().height() - 100 * len(plot_indexes)
        self.widgt.setFixedHeight(height)
        self.draw()

    def hide_curr_plot(self):
        pos = self.curr_plot.get_position()
        h = pos.y1 - pos.y0
        pos.y0 = pos.y1
        self.curr_plot.set_position(pos)
        for i in range(self.curr_plot_index + 1, len(self.axes)):
            pos = self.axes[i].get_position()
            pos.y0 += h
            pos.y1 += h
            self.axes[i].set_position(pos)
        # currently consider every plot height are all 100px height
        height = self.widgt.geometry().height() - 100
        self.widgt.setFixedHeight(height)
        self.hidden_plots[self.curr_plot_index] = h
        self.draw()

    def show_hidden_plot(self, index):
        h = self.hidden_plots[index]
        work_plot = self.axes[index]
        pos = work_plot.get_position()
        pos.y1 = pos.y0 + h
        work_plot.set_position(pos)
        for i in range(index, len(self.axes)):
            pos = self.axes[i].get_position()
            pos.y0 -= h
            pos.y1 -= h
            self.axes[i].set_position(pos)
        # currently consider every plot height are all 100px height
        height = self.widgt.geometry().height() + 100
        self.widgt.setFixedHeight(height)
        del self.hidden_plots[index]
        pos = work_plot.get_position()
        self.draw()

    def show_all_hidden_plots(self):
        plot_indexes = sorted(self.hidden_plots.keys())
        idx = 0
        total_h = 0
        for i in range(plot_indexes[0], len(self.axes)):
            pos = self.axes[i].get_position()
            if idx < len(plot_indexes) and i == plot_indexes[idx]:
                h = self.hidden_plots[i]
                total_h += h
                pos.y1 = pos.y0 + h
                idx += 1
                del self.hidden_plots[i]
            pos.y0 -= total_h
            pos.y1 -= total_h
            self.axes[i].set_position(pos)

        # currently consider every plot height are all 100px height
        height = self.widgt.geometry().height() + 100 * len(plot_indexes)
        self.widgt.setFixedHeight(height)
        self.draw()

    def replot(self):
        for ax in self.axes:
            print("pos", self.ax.get_position())

    # ============================= END AXES ==============================
    def set_colors(self, mode):
        self.c_mode = mode
        self.display_color = set_colors(mode)
        self.fig.patch.set_facecolor(self.display_color['background'])

    def set_peer_plotting_widgets(self, widgets):
        self.peer_plotting_widgets = widgets
