import pathlib
import os
import re
from datetime import datetime
from copy import deepcopy
from PySide2 import QtCore, QtWidgets

from sohstationviewer.view.ui.main_ui import Ui_MainWindow
from sohstationviewer.view.calendar_dialog import CalendarDialog
from sohstationviewer.view.core.filelistwidget import FileListItem
from sohstationviewer.controller.processing import loadData, detectDataType
from sohstationviewer.view.data_type_dialog import DataTypeDialog
from sohstationviewer.view.param_dialog import ParamDialog
from sohstationviewer.view.channel_dialog import ChannelDialog
from sohstationviewer.view.plot_type_dialog import PlotTypeDialog
from sohstationviewer.view.channel_prefer_dialog import ChannelPreferDialog
from sohstationviewer.database.proccessDB import executeDB_dict
from sohstationviewer.view.waveform_dialog import WaveformDialog
from sohstationviewer.view.time_power_squareddialog import (
    TimePowerSquaredDialog)
from sohstationviewer.controller.util import displayTrackingInfo
from sohstationviewer.conf.constants import TM_FM


class MainWindow(QtWidgets.QMainWindow, Ui_MainWindow):

    currentDirectoryChanged = QtCore.Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setupUi(self)

        # Options
        self.dateFormat = 'YYYY-MM-DD'
        self.massPosVoltRangeOpt = 'regular'
        self.bitweightOpt = ''
        self.getChannelPrefer()
        self.YYYY_MM_DDAction.triggered.emit()
        self.waveformDlg = WaveformDialog(self)
        self.TPSDlg = TimePowerSquaredDialog(self)

    @QtCore.Slot()
    def openDataType(self):
        win = DataTypeDialog(self)
        win.show()

    @QtCore.Slot()
    def openParam(self):
        win = ParamDialog(self)
        win.show()

    @QtCore.Slot()
    def openChannel(self):
        win = ChannelDialog(self)
        win.show()

    @QtCore.Slot()
    def openCalendarWidget(self):
        calendar = CalendarDialog(self)
        calendar.show()

    @QtCore.Slot()
    def openPlotType(self):
        win = PlotTypeDialog(self)
        win.show()

    @QtCore.Slot()
    def openChannelPreferences(self):
        dirnames = [os.path.join(self.cwdLineEdit.text(), item.text())
                    for item in self.openFilesList.selectedItems()]
        if dirnames == []:
            msg = "No directories has been selected."
            QtWidgets.QMessageBox.warning(self, "Select directory", msg)
            return
        win = ChannelPreferDialog(self, dirnames)
        win.show()

    @QtCore.Slot()
    def allChanClicked(self):
        if not self.allChanCheckBox.isChecked():
            if self.IDs == []:
                self.allChanCheckBox.setChecked(True)
            else:
                self.currIDsNameLineEdit.setText(self.IDsName)
        else:
            self.currIDsNameLineEdit.setText('')

    @QtCore.Slot()
    def resetView(self):
        self.plottingWidget.reset_view()

    @QtCore.Slot()
    def setDateFormat(self, displayFormat):
        """
        Sets the calendar format used by the QDateEdit text boxes.
        :param displayFormat: str
                A valid display format to be used for date conversion.
        """
        self.timeToDateEdit.setDisplayFormat(displayFormat)
        self.timeFromDateEdit.setDisplayFormat(displayFormat)
        self.dateFormat = displayFormat

    @QtCore.Slot()
    def openFilesListItemDoubleClicked(self, item):
        """
        Handles the double-click event emitted when a user double-clicks on an
        item contained within openFilesList

        Parameters
        ----------
        item : FileListItem
                The item contained within openFilesList which was clicked.
        """
        print(f'Opening {item.text()}')

        # TODO: Do something with the Path object,
        # i.e., path.open(), or path.iterdir() ...

    @QtCore.Slot()
    def changeCurrentDirectory(self):
        """
        Constructs a QFileDialog to select a new working directory
        from which the user can load data. The starting directory
        is taken from cwdLineEdit.
        """
        fd = QtWidgets.QFileDialog(self)
        fd.setFileMode(QtWidgets.QFileDialog.Directory)
        fd.setDirectory(self.cwdLineEdit.text())
        fd.exec_()
        new_path = fd.selectedFiles()[0]
        self.setCurrentDirectory(new_path)

    @QtCore.Slot()
    def readSelectedFiles(self):
        try:
            del self.dataObject
        except Exception:
            pass
        print('Reading currently selected file.')

        # TODO: Launch a separate thread, so that
        # the UI doesn't hang while data is being read.
        # Otherwise, the user won't be able to cancel.

        self.reqSOHChans = (self.IDs if not self.allChanCheckBox.isChecked()
                            else [])

        # TODO: Having a form for user to create the list of channels to draw
        """
        reqSOHChans: list of chans to read data from
        It can be all chans in db or preference list of chans
        For Reftek, the list of channels is fixed => may not need
        """
        self.dirnames = [os.path.join(self.cwdLineEdit.text(), item.text())
                         for item in self.openFilesList.selectedItems()]
        if self.dirnames == []:
            msg = "No directories has been selected."
            QtWidgets.QMessageBox.warning(self, "Select directory", msg)
            return
        self.dataType, _ = detectDataType(self, self.dirnames)
        if self.dataType is None:
            return

        # get reqSOHChans
        if (not self.allChanCheckBox.isChecked() and
                self.dataType != self.IDsDataType):
            msg = (f"DataType detected for the selected data set is "
                   f"{self.dataType} which is different to IDs' "
                   f"{self.IDsDataType}.\n"
                   f"SOHStationViewer will read all data available.\n"
                   f"Do you want to continue?")
            result = QtWidgets.QMessageBox.question(
                self, "Confirmation", msg,
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
            if result == QtWidgets.QMessageBox.No:
                return
            self.allChanCheckBox.setChecked(True)
            self.currIDsNameLineEdit.setText('')
            self.reqSOHChans = []
        self.reqWFChans = []
        if self.dataType == 'RT130':
            reqDSs = []
            for idx, DSCheckbox in enumerate(self.dsCheckBoxes):
                if DSCheckbox.isChecked():
                    reqDSs.append(idx + 1)
            self.reqWFChans = reqDSs
        else:
            if self.wfAllCheckBox.isChecked():
                self.reqWFChans = ['*']
            elif self.wfChansLineEdit.text().strip() != "":
                self.reqWFChans = self.wfChansLineEdit.text().split(",")
        startTmStr = self.timeFromDateEdit.date().toString(QtCore.Qt.ISODate)
        endTmStr = self.timeToDateEdit.date().toString(QtCore.Qt.ISODate)
        self.startTm = datetime.strptime(startTmStr, TM_FM).timestamp()
        self.endTm = datetime.strptime(endTmStr, TM_FM).timestamp()
        self.dataObject = loadData(self.dataType,
                                   self.trackingInfoTextBrowser,
                                   self.dirnames,
                                   reqWFChans=self.reqWFChans,
                                   reqSOHChans=self.reqSOHChans,
                                   readStart=self.startTm,
                                   readEnd=self.endTm)
        self.replotLoadedData()

    @QtCore.Slot()
    def replotLoadedData(self):
        if self.detectGapCheckBox.isChecked():
            self.minGap = self.gapLenLineEdit.text()
        else:
            self.minGap = None
        do = self.dataObject

        timeTickTotal = 5  # TODO: let user choose max ticks to be displayed

        selKey = do.selectedKey

        # mainPlot

        soh_data = deepcopy(do.SOHData[selKey])
        sohChans = list(soh_data.keys())
        mp_data = deepcopy(do.massPosData[selKey])
        if len(self.reqWFChans) != 0:
            wf_data = deepcopy(do.waveformData[selKey]['readData'])
        else:
            wf_data = []
        self.plottingWidget.plot_channels(
            self.startTm, self.endTm, selKey,
            do.dataTime[selKey], do.gaps[selKey], sohChans, timeTickTotal,
            soh_data, mp_data)

        peerPlottingWidgets = [self.plottingWidget]

        if self.tpsCheckBox.isChecked():
            peerPlottingWidgets.append(self.TPSDlg.plotting_widget)
            self.TPSDlg.set_data(self.dataType, ','.join(self.dirnames))
            self.TPSDlg.show()
            wfChans = list(do.waveformData[do.selectedKey].keys())
            self.TPSDlg.plotting_widget.plot_channels(
                self.startTm, self.endTm, selKey,
                do.dataTime[selKey],
                wf_data)
        else:
            self.TPSDlg.hide()

        if self.reqWFChans != []:
            # waveformPlot
            peerPlottingWidgets.append(self.waveformDlg.plotting_widget)
            self.waveformDlg.set_data(self.dataType, ','.join(self.dirnames))
            self.waveformDlg.show()
            wfChans = list(do.waveformData[do.selectedKey].keys())
            self.waveformDlg.plotting_widget.plot_channels(
                self.startTm, self.endTm, selKey,
                do.dataTime[selKey], wfChans, timeTickTotal,
                wf_data, mp_data)
        else:
            self.waveformDlg.hide()

        self.plottingWidget.set_peer_plotting_widgets(peerPlottingWidgets)
        self.waveformDlg.plotting_widget.set_peer_plotting_widgets(
            peerPlottingWidgets)
        self.TPSDlg.plotting_widget.set_peer_plotting_widgets(
            peerPlottingWidgets)

    def setCurrentDirectory(self, path=''):
        # Remove entries when cwd changes
        self.openFilesList.clear()
        # Signal cwd changed, and gather list of files in new cwd
        self.currentDirectoryChanged.emit(path)
        for dent in pathlib.Path(path).iterdir():
            if not dent.is_dir() or dent.name.startswith('.'):
                continue

            self.openFilesList.addItem(FileListItem(dent))

    def getChannelPrefer(self):
        self.IDsName = ''
        self.IDs = []
        self.dataType = 'Unknown'
        rows = executeDB_dict('SELECT name, IDs, dataType FROM ChannelPrefer '
                              'WHERE current=1')
        if len(rows) > 0:
            self.IDsName = rows[0]['name']
            self.IDs = [t.strip() for t in rows[0]['IDs'].split(',')]
            self.IDsDataType = rows[0]['dataType']


def getReqTPSChans(parent, enteredTPSChans, waveformReqChans):
    # TODO: unittest (may need to modify for better filter)
    TPSReqChans = []
    for c in enteredTPSChans:
        TPSRE = re.compile(c.strip().replace('*', '[ZNE123]'))
        reqChans = [chan for chan in waveformReqChans
                    if TPSRE.match(chan)]
        if reqChans == []:
            msg = "%s doesn't match any seismic data channel." % c
            displayTrackingInfo(parent, msg, 'warning')
        else:
            TPSReqChans += reqChans
    return TPSReqChans
