# UI and connectSignals for MainWindow
from math import sqrt
import numpy as np

from PySide2 import QtWidgets, QtCore

from sohstationviewer.view.core import plotting_widget as plottingWidget
from sohstationviewer.controller.plottingData import (
    getTitle, getDayTicks, formatTime)
from sohstationviewer.model.handling_data import (
    get_trimTPSData, get_eachDay5MinList, findTPSTm)
from sohstationviewer.database.extractData import (
    getColorDef, getColorRanges, getChanLabel)
from sohstationviewer.conf.colorSettings import Clr
from sohstationviewer.conf import constants as const
from sohstationviewer.controller.util import displayTrackingInfo, fmti


class TimePowerSquaredWidget(plottingWidget.PlottingWidget):

    def plot_channels(self, start_tm=None, end_tm=None, key=None,
                      data_time=None, plotting_data=None):
        self.processing_log = []  # [(message, type)]

        self.min_x = self.curr_min_x = max(data_time[0], start_tm)
        self.max_x = self.curr_max_x = min(data_time[1], end_tm)
        self.errors = []
        if self.axes != []:
            self.fig.clear()
        self.date_mode = self.parent.date_format.upper()
        if plotting_data is not None:
            self.plotting_data = plotting_data
            self.min_x = max(data_time[0], start_tm)
            self.max_x = min(data_time[1], end_tm)
            self.plot_no = len(plotting_data)
        title = getTitle(key, self.min_x, self.max_x, self.date_mode)
        self.plotting_bot = plottingWidget.BOTTOM
        self.plotting_bot_pixel = plottingWidget.BOTTOM_PX
        self.axes = []
        self.rulers = []
        self.zoom_marker1s = []
        self.zoom_marker2s = []

        self.timestamp_bar_top = self.add_timestamp_bar(0.)
        self.set_title(title, y=0, v_align='bottom')
        self.each_day5_min_list = get_eachDay5MinList(self.min_x, self.max_x)
        for chanID in self.plotting_data:
            ax = self.get_zoom_data(self.plotting_data[chanID], chanID)
            self.axes.append(ax)
        self.set_legend()
        # Set view size fit with the given data
        if self.widgt.geometry().height() < self.plotting_bot_pixel:
            self.widgt.setFixedHeight(self.plotting_bot_pixel)
        self.set_lim_markers()
        self.draw()

    def get_zoom_data(self, c_data, chan_id):
        if 'tps_data' not in c_data:
            # get new minX, maxX according to exact start time  of days
            get_trimTPSData(c_data, self.min_x, self.max_x,
                            self.each_day5_min_list)
        total_days = c_data['tps_data'].shape[0]
        plot_h = self.get_height(1.5 * total_days, bw_plots_distance=0.003)
        ax = self.create_axes(self.plotting_bot, plot_h)
        ax.text(
            -0.1, 1.2,
            f"{getChanLabel(chan_id)} {c_data['samplerate']}",
            horizontalalignment='left',
            verticalalignment='top',
            rotation='horizontal',
            transform=ax.transAxes,
            color=self.display_color['plot_label'],
            size=self.font_size + 2
        )

        zm1 = ax.plot(
            [], [], marker='|', markersize=10,
            markeredgecolor=self.display_color['zoom_marker'])[0]
        self.zoom_marker1s.append(zm1)

        zm2 = ax.plot(
            [], [], marker='|', markersize=10,
            markeredgecolor=self.display_color['zoom_marker'])[0]
        self.zoom_marker2s.append(zm2)

        rl = ax.plot(
            [], [], marker='s', markersize=5,
            markeredgecolor=self.display_color['time_ruler'],
            markerfacecolor='None')[0]
        self.rulers.append(rl)

        x = np.array([i for i in range(const.NO_5M_DAY)])
        s_cnts = self.parent.sel_square_counts  # square counts range
        clrs = self.parent.color_def  # colordef

        for dayIdx, y in enumerate(c_data['tps_data']):
            # not draw data out of day range
            color_set = self.get_color_set(y, s_cnts, clrs)
            # (- dayIdx): each day is a line, increase from top to bottom
            ax.scatter(x, [- dayIdx] * len(x), marker='|',
                       c=color_set, s=7, alpha=0.8)
        # extra to show highlight square
        ax.set_ylim(-(c_data['tps_data'].shape[0] + 1), 1)

        return ax

    def set_legend(self):
        """
        plot one dot for each color and assign label to it
        legend will be one color for each label
        """
        # set height of legend and distance bw legend and upper ax
        plot_h = self.get_height(7, bw_plots_distance=0.003)
        ax = self.canvas.figure.add_axes(
            [self.plotting_l, self.plotting_bot, self.plotting_w, plot_h],
            picker=True
        )
        ax.patch.set_alpha(0)
        c_labels = self.parent.sel_col_labels
        clrs = self.parent.color_def  # colordef
        for idx in range(len(c_labels)):
            # draw a dot out of xlim so it isn't displayed in plotting area
            ax.scatter([300], [1],
                       c=Clr[clrs[idx]],
                       s=0.1,
                       label=c_labels[idx],
                       edgecolor=self.display_color['basic'],
                       alpha=0.8,
                       zorder=1,
                       picker=True)
        # extra to show highlight square
        ax.set_xlim(-2, const.NO_5M_DAY + 1)
        ax.legend(loc="upper left", framealpha=0.2,
                  markerscale=25,
                  labelcolor=self.display_color['basic'])

    def get_color_set(self, y, s_cnts, col):
        """
        create array of color (col) according to value of y compare with
         sCnts (square count range)
        y: value to build color on
        sCnts: list of square count range is the range of square counts
            corresponding to colors (col)
        col: list of color names based on colorSettings.Clr
        """
        return (
            np.where(
                y == s_cnts[0], Clr[col[0]], np.where(
                    y < s_cnts[1], Clr[col[1]], np.where(
                        y < s_cnts[2], Clr[col[2]], np.where(
                            y < s_cnts[3], Clr[col[3]], np.where(
                                y < s_cnts[4], Clr[col[4]], np.where(
                                    y < s_cnts[5], Clr[col[5]], np.where(
                                        y < s_cnts[6], Clr[col[6]], Clr[col[7]]
                                    ))))))))

    def create_axes(self, plot_b, plot_h, has_min_max_lines=False):
        """
        create axes for 288 of 5m in a day in which minor tick for every hour,
        major tick for every 4 hour
        """
        ax = self.canvas.figure.add_axes(
            [self.plotting_l, plot_b, self.plotting_w, plot_h],
            picker=True
        )
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.xaxis.grid(True, which='major',
                      color=self.display_color['basic'], linestyle='-')
        ax.xaxis.grid(True, which='minor',
                      color=self.display_color['sub_basic'], linestyle='-')
        ax.set_yticks([])

        times, major_times, major_time_labels = getDayTicks()
        ax.set_xticks(times, minor=True)
        ax.set_xticks(major_times)
        ax.set_xticklabels(major_time_labels, fontsize=self.font_size,
                           color=self.display_color['basic'])
        # extra to show highlight square
        ax.set_xlim(-2, const.NO_5M_DAY + 1)
        ax.patch.set_alpha(0)
        return ax

    def on_pick_on_artist(self, event):
        """
        Click on each point will highlight time position for each channel
        and display time and counts for each channel
        """
        info_str = ""
        if event.artist in self.axes:
            xdata = round(event.mouseevent.xdata)
            # when click on outside xrange that close to edge, adjust to edge
            if xdata in [-2, -1]:
                xdata = 0
            if xdata in [288, 289]:
                xdata = 287
            ydata = round(event.mouseevent.ydata)
            if xdata is not None:
                y_idx = - ydata
                x_idx = xdata
                # identify time for other plots' rulers displaying
                self.tps_t = self.each_day5_min_list[y_idx, x_idx]
                format_t = formatTime(self.tps_t, self.date_mode, 'HH:MM:SS')
                info_str += f"{format_t}:"
                for chanID in self.plotting_data:
                    c_data = self.plotting_data[chanID]
                    data = c_data['tps_data'][y_idx, x_idx]
                    info_str += f"  {chanID}:{fmti(sqrt(data))}"
                info_str += "  (counts)"
                displayTrackingInfo(self.tracking_box, info_str)
                self.draw()

    def on_ctrl_cmd_click(self, xdata):
        """
        Ctrl + cmd: base on xdata to find indexes of x an y
        in self.each_day_5_min_list to display ruler for each channel
        """
        self.zoom_marker1_shown = False
        x_idx, y_idx = findTPSTm(xdata, self.each_day5_min_list)
        for rl in self.rulers:
            rl.set_data(x_idx, y_idx)

    def on_shift_click(self, xdata):
        """
        Shift + right click in TPS
        on the second of zoommaker, call set_lim_markers to mark the new limit
        """
        if not self.zoom_marker1_shown:
            self.set_rulers_invisible()
            self.curr_min_x = xdata
            self.zoom_marker1_shown = True
        else:
            [self.curr_min_x, self.curr_max_x] = sorted(
                [self.curr_min_x, xdata])
            self.set_lim_markers()
            self.zoom_marker1_shown = False

    def set_rulers_invisible(self):
        for rl in self.rulers:
            rl.set_data([], [])

    def set_lim_markers(self):
        x_idx, y_idx = findTPSTm(self.curr_min_x, self.each_day5_min_list)
        for zm1 in self.zoom_marker1s:
            zm1.set_data(x_idx, y_idx)
            x_idx, y_idx = findTPSTm(self.curr_max_x, self.each_day5_min_list)
        for zm2 in self.zoom_marker2s:
            zm2.set_data(x_idx, y_idx)


class TimePowerSquaredDialog(QtWidgets.QWidget):
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.date_format = self.parent.dateFormat
        self.bitweight_opt = self.parent.bitweightOpt
        self.setGeometry(50, 50, 1200, 700)
        self.setWindowTitle("TPS Plot")

        main_layout = QtWidgets.QVBoxLayout()
        self.setLayout(main_layout)
        main_layout.setContentsMargins(5, 5, 5, 5)
        main_layout.setSpacing(0)

        self.tracking_info_text_browser = QtWidgets.QTextBrowser(self)

        self.plotting_widget = TimePowerSquaredWidget(
            self, self.tracking_info_text_browser, "timepowersquaredwidget")
        main_layout.addWidget(self.plotting_widget, 2)

        bottom_layout = QtWidgets.QHBoxLayout()
        bottom_layout.addSpacing(20)
        main_layout.addLayout(bottom_layout)

        # ################ Color range #################
        bottom_layout.addWidget(QtWidgets.QLabel("Color Range"))

        self.color_def = getColorDef()
        (self.color_ranges,
         self.all_square_counts,
         self.color_label) = getColorRanges()
        self.color_range_choice = QtWidgets.QComboBox(self)
        self.color_range_choice.addItems(self.color_ranges)
        self.color_range_choice.setCurrentText('high')
        bottom_layout.addWidget(self.color_range_choice)
        ################################################
        self.replot_button = QtWidgets.QPushButton("RePlot", self)
        bottom_layout.addWidget(self.replot_button)

        self.write_file_button = QtWidgets.QPushButton('Write file', self)
        bottom_layout.addWidget(self.write_file_button)

        self.tracking_info_text_browser.setFixedHeight(60)
        bottom_layout.addWidget(self.tracking_info_text_browser)

        self.connect_signals()
        self.color_range_changed()

    def set_data(self, data_type, file_name):
        self.data_type = data_type
        self.setWindowTitle("TPS Plot %s - %s" % (data_type, file_name))

    def resizeEvent(self, event):
        self.plotting_widget.init_size()

    def connect_signals(self):
        self.write_file_button.clicked.connect(self.write_file)
        self.replot_button.clicked.connect(self.replot)
        self.color_range_choice.currentTextChanged.connect(
            self.color_range_changed)

    @QtCore.Slot()
    def color_range_changed(self):
        color_range = self.color_range_choice.currentText()
        cr_index = self.color_ranges.index(color_range)
        self.sel_square_counts = self.all_square_counts[cr_index]
        self.sel_col_labels = self.color_label[cr_index]

    @QtCore.Slot()
    def write_file(self):
        print("writeFile")

    @QtCore.Slot()
    def replot(self):
        self.plotting_widget.plot_channels()
