# UI and connectSignals for MainWindow

from PySide2 import QtWidgets

from sohstationviewer.view.core import plotting_widget as plottingWidget
from sohstationviewer.model.handling_data import trim_downsample_WFChan

from sohstationviewer.controller.plottingData import getTitle

from sohstationviewer.database import extractData

from sohstationviewer.conf.dbSettings import dbConf


class WaveformWidget(plottingWidget.PlottingWidget):

    def plot_channels(self, start_tm, end_tm, key,
                      data_time, channel_list, time_ticks_total,
                      plotting_data1, plotting_data2):
        """
        :param setID: (netID, statID, locID)
        :param plottingData: a ditionary including:
            { gaps: [(t1,t2),(t1,t2),...]   (in epoch time)
              channels:[cha: {netID, statID, locID, chanID, times, data,
                              samplerate, startTmEpoch, endTmEpoch}       #
              earliestUTC: the earliest time of all channels
              latestUTC: the latest time of all channels
        :timeTicksTotal: max number of tick to show on time bar

        """
        self.plotting_data1 = plotting_data1
        self.plotting_data2 = plotting_data2
        self.processing_log = []     # [(message, type)]
        self.errors = []
        if self.axes != []:
            self.fig.clear()
        self.date_mode = self.parent.date_format.upper()
        self.time_ticks_total = time_ticks_total
        self.min_x = self.curr_min_x = max(data_time[0], start_tm)
        self.max_x = self.curr_max_x = min(data_time[1], end_tm)
        self.plot_no = len(self.plotting_data1) + len(self.plotting_data2)
        title = getTitle(key, self.min_x, self.max_x, self.date_mode)
        self.plotting_bot = plottingWidget.BOTTOM
        self.plotting_bot_pixel = plottingWidget.BOTTOM_PX
        self.axes = []

        self.timestamp_bar_top = self.add_timestamp_bar(0.003)
        self.set_title(title)

        for chanID in self.plotting_data1:
            chan_db = extractData.getWFPlotInfo(chanID)
            if chan_db['plotType'] == '':
                continue
            self.plotting_data1[chanID]['chan_db'] = chan_db
            self.get_zoom_data(self.plotting_data1[chanID], chanID, True)
        for chanID in self.plotting_data2:
            chan_db = extractData.getChanPlotInfo(chanID,
                                                  self.parent.data_type)
            self.plotting_data2[chanID]['chan_db'] = chan_db
            self.get_zoom_data(self.plotting_data2[chanID], chanID, True)

        self.axes.append(self.plot_none())
        self.timestamp_bar_bottom = self.add_timestamp_bar(0.003, top=False)
        self.set_lim(org_size=True)
        self.bottom = self.axes[-1].get_ybound()[0]
        self.ruler = self.add_ruler(self.display_color['time_ruler'])
        self.zoom_marker1 = self.add_ruler(self.display_color['zoom_marker'])
        self.zoom_marker2 = self.add_ruler(self.display_color['zoom_marker'])
        # Set view size fit with the given data
        if self.widgt.geometry().height() < self.plotting_bot_pixel:
            self.widgt.setFixedHeight(self.plotting_bot_pixel)

        self.draw()

    def get_zoom_data(self, c_data, chan_id, first_time=False):
        """
        :param setID: (netID, statID, locID)
        :param plottingData: a ditionary including:
            { gaps: [(t1,t2),(t1,t2),...]   (in epoch time)
              channels:{cha: {netID, statID, locID, chanID, times, data,
                              samplerate, startTmEpoch, endTmEpoch}       #
              earliestUTC: the earliest time of all channels
              latestUTC: the latest time of all channels
        :timeTicksTotal: max number of tick to show on time bar

        Data set: {channelname: [(x,y), (x,y)...]
        """
        chan_db = c_data['chan_db']
        plot_type = chan_db['plotType']
        # data already processed for massposition in plottingWidget
        if not (chan_id.startswith('VM') or chan_id.startswith('MP')):
            trim_downsample_WFChan(
                c_data, self.curr_min_x, self.curr_max_x, first_time)
            self.apply_convert_factor(c_data, 1)
        # use ax_wf because with massposition, ax has been used
        # in plottingWidget
        if 'ax_wf' not in c_data:
            ax = getattr(self, dbConf['plotFunc'][plot_type][1])(
                c_data, chan_db, chan_id, None, None)
            if ax is None:
                return
            c_data['ax_wf'] = ax
            ax.chan = chan_id
            self.axes.append(ax)
        else:
            getattr(self, dbConf['plotFunc'][plot_type][1])(
                c_data, chan_db, chan_id, c_data['ax_wf'], None)


class WaveformDialog(QtWidgets.QWidget):
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.date_format = self.parent.dateFormat
        self.bitweight_opt = self.parent.bitweightOpt
        self.massPosVoltRangeOpt = self.parent.massPosVoltRangeOpt
        self.setGeometry(300, 300, 1200, 700)
        self.setWindowTitle("Raw Data Plot")

        main_layout = QtWidgets.QVBoxLayout()
        self.setLayout(main_layout)
        main_layout.setContentsMargins(5, 5, 5, 5)
        main_layout.setSpacing(0)

        self.tracking_info_text_browser = QtWidgets.QTextBrowser(self)

        self.plotting_widget = WaveformWidget(
            self, self.tracking_info_text_browser, "waveformWidget")
        main_layout.addWidget(self.plotting_widget, 2)

        bottom_layout = QtWidgets.QHBoxLayout()
        main_layout.addLayout(bottom_layout)

        self.write_ps_button = QtWidgets.QPushButton('Write .ps', self)
        bottom_layout.addWidget(self.write_ps_button)
        self.tracking_info_text_browser.setFixedHeight(60)
        bottom_layout.addWidget(self.tracking_info_text_browser)

        self.connect_signals()

    def set_data(self, data_type, file_name):
        self.data_type = data_type
        self.setWindowTitle("Raw Data Plot %s - %s" % (data_type, file_name))

    def resizeEvent(self, event):
        self.plotting_widget.init_size()

    def connect_signals(self):
        print("connectSignals")
